# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the tests for the helpers.search.models."""
import re
from unittest.mock import MagicMock

import pytest

from aea.exceptions import AEAEnforceError
from aea.helpers.search.models import (
    And,
    Attribute,
    AttributeInconsistencyException,
    Constraint,
    ConstraintExpr,
    ConstraintType,
    ConstraintTypes,
    DataModel,
    Description,
    Location,
    Not,
    Or,
    Query,
    generate_data_model,
)


def test_location():
    """Test Location type."""
    location = Location(latitude=1.1, longitude=2.2)
    assert location is not None

    assert location.tuple == (location.latitude, location.longitude)
    assert location.distance(Location(1.1, 2.2)) == 0
    assert location == Location(1.1, 2.2)
    assert str(location) == "Location(latitude=1.1,longitude=2.2)"

    location_pb = location.encode()
    actual_location = Location.decode(location_pb)
    assert actual_location == location
    assert location.latitude == actual_location.latitude
    assert location.longitude == actual_location.longitude


def test_attribute():
    """Test data model Attribute."""
    params = dict(name="test", type_=str, is_required=True)
    attribute = Attribute(**params)
    assert attribute is not None

    assert attribute == Attribute(**params)
    assert attribute != Attribute(name="another", type_=int, is_required=True)

    assert str(attribute) == "Attribute(name=test,type=<class 'str'>,is_required=True)"

    attribute_pb = attribute.encode()
    actual_attribute = Attribute.decode(attribute_pb)
    assert actual_attribute == attribute


def test_data_model():
    """Test data model definitions."""
    params = dict(name="test", type_=str, is_required=True)
    data_model = DataModel("test", [Attribute(**params)])

    data_model._check_validity()
    with pytest.raises(
        ValueError,
        match="Invalid input value for type 'DataModel': duplicated attribute name.",
    ):
        data_model = DataModel("test", [Attribute(**params), Attribute(**params)])
        data_model._check_validity()

    assert data_model == DataModel("test", [Attribute(**params)])
    assert data_model != DataModel("not test", [Attribute(**params)])

    assert (
        str(data_model)
        == "DataModel(name=test,attributes={'test': \"Attribute(name=test,type=<class 'str'>,is_required=True)\"},description=)"
    )

    data_model_pb = data_model.encode()
    actual_data_model = DataModel.decode(data_model_pb)
    assert actual_data_model == data_model


def test_generate_data_model():
    """Test model generated from description."""
    params = dict(name="test", type_=str, is_required=True)

    data_model = DataModel("test", [Attribute(**params)])

    assert generate_data_model("test", {"test": "str"}) == data_model


def test_description():
    """Test model description."""
    values = {
        "test": "test_value",
        "bool": True,
        "float": 1.1,
        "int": int(1),
        "loc": Location(1, 1),
    }
    description = Description(
        values=values, data_model=generate_data_model("test", values)
    )

    assert description.values == values
    assert description == Description(
        values=values, data_model=generate_data_model("test", values)
    )
    assert list(description.values.values()) == list(values.values())
    assert list(description) == list(values)

    with pytest.raises(
        AttributeInconsistencyException, match=r"Missing required attribute."
    ):
        Description(
            values=values, data_model=generate_data_model("test", {"extra_key": "key"})
        )
    with pytest.raises(
        AttributeInconsistencyException,
        match=r"Have extra attribute not in data model.",
    ):
        Description(values=values, data_model=generate_data_model("test", {}))
    with pytest.raises(
        AttributeInconsistencyException, match=r".* has incorrect type:.*"
    ):
        Description(
            values=values,
            data_model=generate_data_model("test", {**values, "test": 12}),
        )

    with pytest.raises(
        AttributeInconsistencyException, match=r".* has unallowed type:.*"
    ):
        Description(
            values={"test": object()},
            data_model=generate_data_model("test", {"test": object()}),
        )

    assert re.match(r"Description\(values=.*data_model=.*", str(description))

    description_pb = description._encode()
    actual_description = Description._decode(description_pb)
    assert actual_description == description

    mock = MagicMock()
    mock.description_bytes = None
    Description.encode(mock, description)
    assert mock.description_bytes is not None
    description = Description.decode(mock)
    assert "test" in description.values


def test_constraint_type():
    """Test ConstraintType."""
    constraint_type_values = {
        "int": 12,
        "bool": True,
        "float": 10.4,
        "str": "some_string",
        "location": Location(1.1, 2.2),
    }
    constraint_type_types = {
        "int": int,
        "bool": bool,
        "float": float,
        "str": str,
        "location": Location,
    }
    to_check = {
        "int": 13,
        "bool": False,
        "float": 9.3,
        "str": "some_other_string",
        "location": Location(1.2, 2.3),
    }

    # = and !=
    for constraint_types_type in [ConstraintTypes.EQUAL, ConstraintTypes.NOT_EQUAL]:
        for allowed_type in ["int", "bool", "float", "str"]:
            constraint_type = ConstraintType(
                constraint_types_type, constraint_type_values[allowed_type]
            )
            constraint_type.is_valid(
                Attribute("test", constraint_type_types[allowed_type], True)
            )
            constraint_type.check(to_check[allowed_type])
            assert constraint_type == ConstraintType(
                constraint_types_type, constraint_type_values[allowed_type]
            )
            assert (
                str(constraint_type)
                == f"ConstraintType(value={constraint_type_values[allowed_type]},type={constraint_types_type})"
            )

            constraint_type_pb = constraint_type.encode()
            actual_constraint_type = ConstraintType.decode(
                constraint_type_pb, "relation"
            )
            assert actual_constraint_type == constraint_type

    # < and <= and > and >=
    for constraint_types_type in [
        ConstraintTypes.LESS_THAN,
        ConstraintTypes.LESS_THAN_EQ,
        ConstraintTypes.GREATER_THAN,
        ConstraintTypes.GREATER_THAN_EQ,
    ]:
        for allowed_type in ["int", "float", "str"]:
            constraint_type = ConstraintType(
                constraint_types_type, constraint_type_values[allowed_type]
            )
            constraint_type.is_valid(
                Attribute("test", constraint_type_types[allowed_type], True)
            )
            constraint_type.check(to_check[allowed_type])
            assert constraint_type == ConstraintType(
                constraint_types_type, constraint_type_values[allowed_type]
            )
            assert (
                str(constraint_type)
                == f"ConstraintType(value={constraint_type_values[allowed_type]},type={constraint_types_type})"
            )

            constraint_type_pb = constraint_type.encode()
            actual_constraint_type = ConstraintType.decode(
                constraint_type_pb, "relation"
            )
            assert actual_constraint_type == constraint_type

    # within
    constraint_type_values = {
        "int": (1, 2),
        "float": (2.4, 5.4),
        "str": ("str_1", "str_2"),
        "location": (Location(1.1, 2.2), Location(1.2, 5.2)),
    }
    constraint_type_types = {
        "int": int,
        "float": float,
        "str": str,
        "location": Location,
    }
    to_check = {
        "int": 13,
        "float": 9.3,
        "str": "some_other_string",
        "location": Location(1.2, 2.3),
    }

    for range_constraint_type in ["int", "float", "str"]:  # location is not working
        constraint_type = ConstraintType(
            ConstraintTypes.WITHIN, constraint_type_values[range_constraint_type]
        )
        constraint_type.is_valid(
            Attribute("test", constraint_type_types[range_constraint_type], True)
        )
        constraint_type.check(to_check[range_constraint_type])
        assert constraint_type == ConstraintType(
            ConstraintTypes.WITHIN, constraint_type_values[range_constraint_type]
        )
        assert (
            str(constraint_type)
            == f"ConstraintType(value={constraint_type_values[range_constraint_type]},type=within)"
        )
        constraint_type_pb = constraint_type.encode()
        actual_constraint_type = ConstraintType.decode(constraint_type_pb, "range")
        assert actual_constraint_type == constraint_type

    # in and not_in
    constraint_type_values = {
        "int": (1, 2),
        "bool": (True, False),
        "float": (2.4, 5.4),
        "str": ("str_1", "str_2"),
        "location": (Location(1.1, 2.2), Location(1.2, 5.2)),
    }
    constraint_type_types = {
        "int": int,
        "bool": bool,
        "float": float,
        "str": str,
        "location": Location,
    }
    to_check = {
        "int": 13,
        "bool": False,
        "float": 9.3,
        "str": "some_other_string",
        "location": Location(1.2, 2.3),
    }

    for constraint_types_type in [ConstraintTypes.IN, ConstraintTypes.NOT_IN]:
        for constraint_set in ["int", "bool", "float", "str", "location"]:
            constraint_type = ConstraintType(
                constraint_types_type, constraint_type_values[constraint_set]
            )
            constraint_type.is_valid(
                Attribute("test", constraint_type_types[constraint_set], True)
            )
            constraint_type.check(to_check[constraint_set])
            assert constraint_type == ConstraintType(
                constraint_types_type, constraint_type_values[constraint_set]
            )
            assert (
                str(constraint_type)
                == f"ConstraintType(value={constraint_type_values[constraint_set]},type={constraint_types_type})"
            )
            constraint_type_pb = constraint_type.encode()
            actual_constraint_type = ConstraintType.decode(constraint_type_pb, "set")
            assert actual_constraint_type == constraint_type

    # distance
    constraint_location = (Location(1.1, 2.2), 2.2)
    constraint_type_distance = ConstraintType(
        ConstraintTypes.DISTANCE, constraint_location
    )
    constraint_type_distance.is_valid(Attribute("test", int, True))
    constraint_type_distance.check(Location(1.1, 2.2))
    assert constraint_type_distance == ConstraintType(
        ConstraintTypes.DISTANCE, constraint_location
    )
    constraint_type_distance_pb = constraint_type_distance.encode()
    actual_constraint_type_distance = ConstraintType.decode(
        constraint_type_distance_pb, "distance"
    )
    assert actual_constraint_type_distance == constraint_type_distance

    # failures
    with pytest.raises(ValueError):
        ConstraintType("something", [Location(1.1, 2.2), 2.2]).is_valid(
            Attribute("test", int, True)
        )

    with pytest.raises(AEAEnforceError, match=""):
        ConstraintType(ConstraintTypes.GREATER_THAN, str)

    list_value = [1, 2]
    set_value = {1, 2}
    list_location = [Location(1.1, 2.2), 2.2]

    with pytest.raises(
        AEAEnforceError, match=f"Expected tuple, got {type(list_value)}"
    ):
        ConstraintType(ConstraintTypes.WITHIN, list_value)

    with pytest.raises(
        AEAEnforceError, match=f"Expected tuple, got {type(list_value)}"
    ):
        ConstraintType(ConstraintTypes.IN, list_value)

    with pytest.raises(AEAEnforceError, match=f"Expected tuple, got {type(set_value)}"):
        ConstraintType(ConstraintTypes.IN, set_value)

    with pytest.raises(
        AEAEnforceError, match=f"Expected tuple, got {type(list_value)}"
    ):
        ConstraintType(ConstraintTypes.NOT_IN, list_value)

    with pytest.raises(AEAEnforceError, match=f"Expected tuple, got {type(set_value)}"):
        ConstraintType(ConstraintTypes.NOT_IN, set_value)

    with pytest.raises(
        AEAEnforceError, match=f"Expected tuple, got {type(list_location)}"
    ):
        ConstraintType(ConstraintTypes.DISTANCE, list_location)

    incorrect_category = "some_incorrect_category"
    with pytest.raises(
        ValueError,
        match=r"Incorrect category. Expected either of .* Found some_incorrect_category.",
    ):
        constraint_type_distance_pb = constraint_type_distance.encode()
        ConstraintType.decode(constraint_type_distance_pb, incorrect_category)


def test_constraints_expression():
    """Test constraint expressions: And, Or, Not, Constraint."""
    and_expression = And(
        [
            Constraint("number", ConstraintType(ConstraintTypes.LESS_THAN, 15)),
            Constraint("number", ConstraintType(ConstraintTypes.GREATER_THAN, 10)),
        ]
    )
    and_expression.check_validity()
    assert and_expression.check(Description({"number": 12}))
    assert and_expression.is_valid(
        DataModel("some_name", [Attribute("number", int, True)])
    )
    and_expression_pb = ConstraintExpr._encode(and_expression)
    actual_and_expression = ConstraintExpr._decode(and_expression_pb)
    assert actual_and_expression == and_expression

    or_expression = Or(
        [
            Constraint("number", ConstraintType(ConstraintTypes.EQUAL, 12)),
            Constraint("number", ConstraintType(ConstraintTypes.EQUAL, 13)),
        ]
    )
    or_expression.check_validity()
    assert or_expression.check(Description({"number": 12}))
    assert or_expression.is_valid(
        DataModel("some_name", [Attribute("number", int, True)])
    )
    or_expression_pb = ConstraintExpr._encode(or_expression)
    actual_or_expression = ConstraintExpr._decode(or_expression_pb)
    assert actual_or_expression == or_expression

    not_expression = Not(
        And(
            [
                Constraint("number", ConstraintType(ConstraintTypes.EQUAL, 12)),
                Constraint("number", ConstraintType(ConstraintTypes.EQUAL, 12)),
            ]
        )
    )
    not_expression.check_validity()
    assert not_expression.check(Description({"number": 13}))
    assert not_expression.is_valid(
        DataModel("some_name", [Attribute("number", int, True)])
    )
    not_expression_pb = ConstraintExpr._encode(not_expression)
    actual_not_expression = ConstraintExpr._decode(not_expression_pb)
    assert actual_not_expression == not_expression

    # constraint
    constraint_expression = Constraint("author", ConstraintType("==", "Stephen King"))
    constraint_expression.check_validity()
    assert constraint_expression.check(Description({"author": "Stephen King"}))
    assert constraint_expression.is_valid(
        DataModel("some_name", [Attribute("author", str, True)])
    )
    constraint_expression_pb = ConstraintExpr._encode(constraint_expression)
    actual_constraint_expression = ConstraintExpr._decode(constraint_expression_pb)
    assert actual_constraint_expression == constraint_expression

    incorrect_expression = Location(1.1, 2.2)
    with pytest.raises(
        ValueError,
        match=f"Invalid expression type. Expected either of 'And', 'Or', 'Not', 'Constraint'. Found {type(incorrect_expression)}.",
    ):
        ConstraintExpr._encode(incorrect_expression)


def test_constraints_and():
    """Test And."""
    and_expression = And(
        [
            Constraint("number", ConstraintType(ConstraintTypes.LESS_THAN, 15)),
            Constraint("number", ConstraintType(ConstraintTypes.GREATER_THAN, 10)),
        ]
    )
    and_expression.check_validity()
    assert and_expression.check(Description({"number": 12}))
    assert and_expression.is_valid(
        DataModel("some_name", [Attribute("number", int, True)])
    )
    and_expression_pb = and_expression.encode()
    actual_and_expression = And.decode(and_expression_pb)
    assert actual_and_expression == and_expression


def test_constraints_or():
    """Test Or."""
    or_expression = Or(
        [
            Constraint("number", ConstraintType(ConstraintTypes.EQUAL, 12)),
            Constraint("number", ConstraintType(ConstraintTypes.EQUAL, 13)),
        ]
    )
    or_expression.check_validity()
    assert or_expression.check(Description({"number": 12}))
    assert or_expression.is_valid(
        DataModel("some_name", [Attribute("number", int, True)])
    )
    or_expression_pb = or_expression.encode()
    actual_or_expression = Or.decode(or_expression_pb)
    assert actual_or_expression == or_expression


def test_constraints_not():
    """Test Not."""
    not_expression = Not(
        And(
            [
                Constraint("number", ConstraintType(ConstraintTypes.EQUAL, 12)),
                Constraint("number", ConstraintType(ConstraintTypes.EQUAL, 12)),
            ]
        )
    )
    not_expression.check_validity()
    assert not_expression.check(Description({"number": 13}))
    assert not_expression.is_valid(
        DataModel("some_name", [Attribute("number", int, True)])
    )
    not_expression_pb = not_expression.encode()
    actual_not_expression = Not.decode(not_expression_pb)
    assert actual_not_expression == not_expression


def test_constraint():
    """Test Constraint."""
    c1 = Constraint("author", ConstraintType("==", "Stephen King"))
    c2 = Constraint("year", ConstraintType("within", (2000, 2010)))
    c3 = Constraint("author", ConstraintType("in", ("Stephen King", "J. K. Rowling")))
    c4 = Constraint(
        "author", ConstraintType("not_in", ("Stephen King", "J. K. Rowling"))
    )
    c5 = Constraint("address", ConstraintType("distance", (Location(1.1, 2.2), 2.2)))

    book_1 = Description(
        {"author": "Stephen King", "year": 2005, "address": Location(1.1, 2.2)}
    )
    book_2 = Description(
        {"author": "George Orwell", "year": 1948, "address": Location(1.1, 2.2)}
    )

    assert c1.check(book_1)
    assert not c1.check(book_2)
    # empty description
    assert not c1.check(Description({}))
    # bad type
    assert not c1.check(Description({"author": 12}))
    # bad type
    assert not c2.check(Description({"author": 12}))

    assert c1.is_valid(generate_data_model("test", {"author": "some author"}))
    assert not c1.is_valid(generate_data_model("test", {"not_author": "some author"}))

    assert c1 == c1
    assert c1 != c2

    assert (
        str(c1)
        == f"Constraint(attribute_name=author,constraint_type={c1.constraint_type})"
    )
    assert (
        str(c2)
        == f"Constraint(attribute_name=year,constraint_type={c2.constraint_type})"
    )

    c1_pb = c1.encode()
    actual_c1 = Constraint.decode(c1_pb)
    assert actual_c1 == c1

    c2_pb = c2.encode()
    actual_c2 = Constraint.decode(c2_pb)
    assert actual_c2 == c2

    c3_pb = c3.encode()
    actual_c3 = Constraint.decode(c3_pb)
    assert actual_c3 == c3

    c4_pb = c4.encode()
    actual_c4 = Constraint.decode(c4_pb)
    assert actual_c4 == c4

    c5_pb = c5.encode()
    actual_c5 = Constraint.decode(c5_pb)
    assert actual_c5 == c5


def test_query():
    """Test Query."""
    c1 = Constraint("author", ConstraintType("==", "Stephen King"))
    model = generate_data_model("book_author", {"author": "author of the book"})
    query = Query([c1], model=model)

    assert query.check(
        Description({"author": "Stephen King", "year": 1991, "genre": "horror"})
    )
    assert query.is_valid(generate_data_model("test", {"author": "some author"}))

    query.check_validity()
    with pytest.raises(ValueError, match=r"Constraints must be a list .*"):
        query = Query(c1)
    Query([]).check_validity()
    with pytest.raises(
        ValueError,
        match=r"Invalid input value for type 'Query': the query is not valid for the given data model.",
    ):
        Query(
            [c1], generate_data_model("test", {"notauthor": "not some author"})
        ).check_validity()

    assert query == Query(
        [c1], model=generate_data_model("book_author", {"author": "author of the book"})
    )

    assert (
        str(query)
        == f"Query(constraints=['Constraint(attribute_name=author,constraint_type=ConstraintType(value=Stephen King,type===))'],model={model})"
    )

    query_pb = query._encode()
    actual_query = Query._decode(query_pb)
    assert actual_query == query

    query_pb = MagicMock()
    query_pb.query_bytes = None
    Query.encode(query_pb, query)
    assert query_pb.query_bytes is not None
    query = Query.decode(query_pb)
    assert "author" in query.model.attributes_by_name



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Check aea dependencies."""
import copy
import dis
import importlib
import re
import sys
from collections import defaultdict
from functools import wraps
from pathlib import Path
from typing import Any, Callable, Dict, Generator, List, Optional, Set, Tuple, Union

from pip._internal.commands.show import search_packages_info  # type: ignore


AEA_ROOT_DIR = Path(__file__).parent.parent

sys.path.append(str(AEA_ROOT_DIR))

from aea.crypto.registries import (  # noqa # pylint: disable=wrong-import-position
    crypto_registry,
)


IGNORE: Set[str] = {"pkg_resources"}
DEP_NAME_RE = re.compile(r"(^[^=><\[]+)", re.I)  # type: ignore


def list_decorator(fn: Callable) -> Callable:
    """Wraps generator to return list."""

    @wraps(fn)
    def wrapper(*args: Any, **kwargs: Any) -> List[Any]:
        return list(fn(*args, **kwargs))

    return wrapper


class DependenciesTool:
    """Tool to work with setup.py dependencies."""

    @staticmethod
    def get_package_files(package_name: str) -> List[str]:
        """Get package files list."""
        packages_info = list(search_packages_info([package_name]))
        if len(packages_info) == 0:
            raise Exception(f"package {package_name} not found")
        files = packages_info[0]["files"]
        location = packages_info[0]["location"]
        return [str(Path(location) / i) for i in files]

    @staticmethod
    def clean_dependency_name(dependecy_specification: str) -> str:
        """Get dependency name from dependency specification."""
        match = DEP_NAME_RE.match(dependecy_specification)
        if not match:
            raise ValueError(f"Bad dependency specification: {dependecy_specification}")
        return match.groups()[0]


class ImportsTool:
    """Tool to work with 3rd part imports in source code."""

    @staticmethod
    def get_imports_for_file(pyfile: Union[str, Path]) -> List[str]:
        """Get all imported modules for python source file."""
        with open(pyfile, "r") as f:
            statements = f.read()
        instructions = dis.get_instructions(statements)  # type: ignore
        imports = [i for i in instructions if "IMPORT" in i.opname]

        grouped: Dict[str, List[str]] = defaultdict(list)
        for instr in imports:
            grouped[instr.opname].append(instr.argval)

        return grouped["IMPORT_NAME"]

    @staticmethod
    def get_module_file(module_name: str) -> str:
        """Get module source file name."""
        try:
            mod = importlib.import_module(module_name)
            return getattr(mod, "__file__", "")
        except (AttributeError, ModuleNotFoundError):
            return ""

    @staticmethod
    @list_decorator
    def list_all_pyfiles(
        root_path: Union[Path, str], pattern: str = "**/*.py"
    ) -> Generator:
        """List all python files in directory."""
        root_path = Path(root_path)
        for path in root_path.rglob(pattern):
            yield path.relative_to(root_path)

    @classmethod
    @list_decorator
    def get_third_part_imports_for_file(cls, pyfile: str) -> Generator:
        """Get list of third part modules imported for source file."""
        imports = cls.get_imports_for_file(pyfile)
        for module_name in imports:
            pyfile = cls.get_module_file(module_name)
            if not pyfile:
                continue
            if "site-packages" not in Path(pyfile).parts:
                continue
            yield module_name, pyfile

    @classmethod
    @list_decorator
    def list_all_pyfiles_with_3rdpart_imports(
        cls, root_path: Union[str, Path], pattern: str = "**/*.py"
    ) -> Generator:
        """Get list of all python sources with 3rd party modules imported."""
        for pyfile in cls.list_all_pyfiles(root_path, pattern=pattern):
            mods = list(cls.get_third_part_imports_for_file(root_path / pyfile))
            if not mods:
                continue
            yield str(pyfile), list(set(mods))


class CheckTool:
    """Tool to check imports in sources match dependencies in setup.py."""

    @classmethod
    def get_section_dependencies_from_setup(cls) -> Dict[str, Dict[str, List[str]]]:
        """Get sections with dependencies with files lists."""
        spec = importlib.util.spec_from_file_location(
            "setup", str(AEA_ROOT_DIR / "setup.py")
        )
        setup = importlib.util.module_from_spec(spec)
        sys.modules[spec.name] = setup
        spec.loader.exec_module(setup)  # type: ignore

        sections_dependencies = copy.deepcopy(setup.all_extras)  # type: ignore
        sections_dependencies.pop("all")
        base = setup.base_deps  # type: ignore

        for crypto_id in crypto_registry.supported_ids:  # type: ignore
            if crypto_id == "fetchai":
                crypto_id = "fetch"
            sections_dependencies.pop(crypto_id)
        sections_dependencies["base"] = base

        return cls.sections_dependencies_add_files(sections_dependencies)

    @staticmethod
    def sections_dependencies_add_files(
        sections_dependencies: Dict[str, List[str]]
    ) -> Dict[str, Dict[str, List[str]]]:
        """Add packages file lists to dependencies in sections."""
        result: Dict[str, Dict[str, List[str]]] = defaultdict(lambda: defaultdict(list))
        for section, deps in sections_dependencies.items():
            for dep in deps:
                dep = DependenciesTool.clean_dependency_name(dep)
                result[section][dep] = DependenciesTool.get_package_files(dep)
        return result

    @classmethod
    def run(cls) -> None:
        """Run dependency check."""
        print("Dependencies check report:")
        print("==========================")
        files_and_modules = ImportsTool.list_all_pyfiles_with_3rdpart_imports(
            AEA_ROOT_DIR, pattern="aea/**/*.py"
        )

        sections_dependencies = cls.get_section_dependencies_from_setup()
        sections_imports = cls.make_sections_with_3rdpart_imports(
            files_and_modules, set(sections_dependencies.keys())
        )
        missed_deps_for_imports, deps_not_imported_directly = cls.check_imports(
            sections_imports, sections_dependencies  # type: ignore
        )

        for section, unresolved_imports in missed_deps_for_imports.items():
            print(
                f"Section `{section}` unresolved imports: {', '.join(unresolved_imports)}"
            )

        if deps_not_imported_directly:
            if missed_deps_for_imports:
                print()
            print(
                f"Dependencies not imported in code directly: {', '.join(deps_not_imported_directly)}"
            )

        if missed_deps_for_imports or deps_not_imported_directly:
            sys.exit(1)
        else:
            print("All good!")

    @staticmethod
    def make_sections_with_3rdpart_imports(
        files_and_modules: List[Tuple[str, List[Tuple[str, str]]]],
        section_names: Set[str],
    ) -> Dict[str, Set[Tuple[str, str]]]:
        """Make sections with list of 3r part imports."""
        sections_imports: Dict[str, Set[Tuple[str, str]]] = defaultdict(set)

        for pyfile, imports in files_and_modules:
            section_name = Path(pyfile).parts[1]

            if section_name not in section_names:
                section_name = "base"
            sections_imports[section_name].update(imports)

        return sections_imports

    @staticmethod
    def check_imports(
        sections_imports: Dict[str, Set[str]],
        sections_dependencies: Dict[str, Dict[str, List[str]]],
    ) -> Tuple[Dict[str, List[str]], List[str]]:
        """Find missing dependencies for imports and not imported dependencies."""

        def _find_dependency_for_module(
            dependencies: Dict[str, List[str]], pyfile: str
        ) -> Optional[str]:
            for package, files in dependencies.items():
                if pyfile in files:
                    return package
            return None

        sections_imports_packages: Dict[str, Dict[str, Optional[str]]] = defaultdict(
            dict
        )
        for section, modules in sections_imports.items():
            for module, pyfile in modules:
                package = _find_dependency_for_module(
                    sections_dependencies.get(section, {}), pyfile
                )
                if module not in IGNORE:
                    sections_imports_packages[section][module] = package

        all_dependencies_set = set(
            sum((list(i.keys()) for _, i in sections_dependencies.items()), [])
        )
        used_dependencies_set = set(
            sum(
                [
                    list(section.values())
                    for section in sections_imports_packages.values()
                ],
                [],
            )
        )
        deps_not_imported_directly: List[str] = list(
            all_dependencies_set - used_dependencies_set
        )
        missed_deps_for_imports: Dict[str, List[str]] = {
            section: [k for k, v in modules.items() if v is None]
            for section, modules in sections_imports_packages.items()
            if None in modules.values()
        }
        return missed_deps_for_imports, deps_not_imported_directly


if __name__ == "__main__":
    CheckTool.run()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2019 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This module contains the tests of the handler classes of the tac negotiation skill."""

import logging
from pathlib import Path
from typing import Optional, cast
from unittest.mock import PropertyMock, patch

import pytest

from aea.crypto.ledger_apis import LedgerApis
from aea.exceptions import AEAEnforceError
from aea.helpers.search.models import (
    Attribute,
    Constraint,
    ConstraintType,
    DataModel,
    Description,
    Query,
)
from aea.helpers.transaction.base import (
    RawMessage,
    RawTransaction,
    SignedTransaction,
    State,
    Terms,
    TransactionDigest,
    TransactionReceipt,
)
from aea.protocols.dialogue.base import DialogueMessage, DialogueStats
from aea.test_tools.test_skill import BaseSkillTestCase, COUNTERPARTY_AGENT_ADDRESS

from packages.fetchai.protocols.contract_api.custom_types import Kwargs
from packages.fetchai.protocols.contract_api.message import ContractApiMessage
from packages.fetchai.protocols.default.message import DefaultMessage
from packages.fetchai.protocols.fipa.message import FipaMessage
from packages.fetchai.protocols.ledger_api.message import LedgerApiMessage
from packages.fetchai.protocols.oef_search.message import OefSearchMessage
from packages.fetchai.protocols.signing.message import SigningMessage
from packages.fetchai.skills.tac_negotiation.dialogues import (
    ContractApiDialogue,
    ContractApiDialogues,
    FipaDialogue,
    FipaDialogues,
    LedgerApiDialogue,
    LedgerApiDialogues,
    OefSearchDialogue,
    OefSearchDialogues,
    SigningDialogue,
    SigningDialogues,
)
from packages.fetchai.skills.tac_negotiation.handlers import (
    ContractApiHandler,
    FipaNegotiationHandler,
    LEDGER_API_ADDRESS,
    LedgerApiHandler,
    OefSearchHandler,
    SigningHandler,
)
from packages.fetchai.skills.tac_negotiation.helpers import SUPPLY_DATAMODEL_NAME
from packages.fetchai.skills.tac_negotiation.strategy import Strategy
from packages.fetchai.skills.tac_negotiation.transactions import Transactions

from tests.conftest import ROOT_DIR


class TestFipaHandler(BaseSkillTestCase):
    """Test fipa handler of generic buyer."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "tac_negotiation")

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.fipa_handler = cast(
            FipaNegotiationHandler, cls._skill.skill_context.handlers.fipa
        )
        cls.strategy = cast(Strategy, cls._skill.skill_context.strategy)
        cls.transactions = cast(Transactions, cls._skill.skill_context.transactions)
        cls.logger = cls._skill.skill_context.logger

        cls.fipa_dialogues = cast(
            FipaDialogues, cls._skill.skill_context.fipa_dialogues
        )
        cls.contract_api_dialogues = cast(
            ContractApiDialogues, cls._skill.skill_context.contract_api_dialogues
        )
        cls.signing_dialogues = cast(
            SigningDialogues, cls._skill.skill_context.signing_dialogues
        )

        cls.dialogue_stats = cls.fipa_dialogues.dialogue_stats
        cls.ledger_id = "some_ledger_id"
        cls.counterprty_address = COUNTERPARTY_AGENT_ADDRESS
        cls.amount_by_currency_id = {"1": 50}
        cls.quantities_by_good_id = {"2": -10}
        cls.nonce = "some_nonce"
        cls.contract_id = "some_contract_id"
        cls.contract_address = "some_contract_address"
        cls.kwargs = {"some_key": "some_value"}
        cls.counterparty_signature = "some_counterparty_signature"
        cls.terms = Terms(
            cls.ledger_id,
            cls._skill.skill_context.agent_address,
            cls.counterprty_address,
            cls.amount_by_currency_id,
            cls.quantities_by_good_id,
            cls.nonce,
        )
        cls.raw_message = RawMessage(
            ledger_id=cls.ledger_id, body=cls.terms.sender_hash.encode("utf-8")
        )

        cls.cfp_query = Query(
            [Constraint("some_attribute", ConstraintType("==", "some_service"))],
            DataModel(
                SUPPLY_DATAMODEL_NAME,
                [
                    Attribute(
                        "some_attribute", str, False, "Some attribute descriptions."
                    )
                ],
            ),
        )
        cls.proposal = Description(
            {
                "ledger_id": cls.ledger_id,
                "price": 100,
                "currency_id": "1",
                "fee": 1,
                "nonce": cls.nonce,
            }
        )
        cls.list_of_messages_other_initiated = (
            DialogueMessage(
                FipaMessage.Performative.CFP, {"query": cls.cfp_query}, True
            ),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE, {"proposal": cls.proposal}
            ),
            DialogueMessage(FipaMessage.Performative.ACCEPT),
            DialogueMessage(
                FipaMessage.Performative.MATCH_ACCEPT_W_INFORM,
                {"info": {"address": "some_term_sender_address"}},
            ),
            DialogueMessage(
                FipaMessage.Performative.INFORM,
                {"info": {"transaction_digest": "some_transaction_digest_body"}},
            ),
        )
        cls.list_of_messages_self_initiated = (
            DialogueMessage(FipaMessage.Performative.CFP, {"query": cls.cfp_query}),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE, {"proposal": cls.proposal}
            ),
            DialogueMessage(FipaMessage.Performative.ACCEPT),
            DialogueMessage(
                FipaMessage.Performative.MATCH_ACCEPT_W_INFORM,
                {"info": {"address": "some_term_sender_address"}},
            ),
            DialogueMessage(
                FipaMessage.Performative.INFORM,
                {"info": {"transaction_digest": "some_transaction_digest_body"}},
            ),
        )

    def test_setup(self):
        """Test the setup method of the fipa handler."""
        assert self.fipa_handler.setup() is None
        self.assert_quantity_in_outbox(0)

    def test_handle_unidentified_dialogue(self):
        """Test the _handle_unidentified_dialogue method of the fipa handler."""
        # setup
        incorrect_dialogue_reference = ("", "")
        incoming_message = self.build_incoming_message(
            message_type=FipaMessage,
            dialogue_reference=incorrect_dialogue_reference,
            performative=FipaMessage.Performative.ACCEPT,
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.fipa_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received invalid fipa message={incoming_message}, unidentified dialogue.",
        )

        self.assert_quantity_in_outbox(1)
        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_outbox(),
            message_type=DefaultMessage,
            performative=DefaultMessage.Performative.ERROR,
            to=incoming_message.sender,
            sender=self.skill.skill_context.agent_address,
            error_code=DefaultMessage.ErrorCode.INVALID_DIALOGUE,
            error_msg="Invalid dialogue.",
            error_data={"fipa_message": incoming_message.encode()},
        )
        assert has_attributes, error_str

    def _assert_stat_state(
        self,
        dialogue_stats: DialogueStats,
        changed_agent: Optional[str] = None,
        changed_end_state: Optional[FipaDialogue.EndState] = None,
    ) -> None:
        """
        Evaluates the state of dialogue stats.

        If 'changed_agent' and 'changed_end_state' are None,
        it asserts that the dialogue stats are 0 for all end_states.

        If they are not None, it checks that all end_states are 0, except for 'changed_end_state'
        in dialogues started by 'changed_agent' (i.e. 'self' or 'other').

        :param changed_agent: can either by 'self' or 'other'. Dialogues started by this agent has a none-zero end_state.
        :param changed_end_state: the changed end_state.
        :return:
        """
        if changed_agent is None and changed_end_state is None:
            unchanged_dict_1 = dialogue_stats.self_initiated
            unchanged_dict_2 = dialogue_stats.other_initiated
            for end_state_numbers in unchanged_dict_1.values():
                assert end_state_numbers == 0
            for end_state_numbers in unchanged_dict_2.values():
                assert end_state_numbers == 0
        elif changed_agent is not None and changed_end_state is not None:
            if changed_agent == "self":
                changed_dict = dialogue_stats.self_initiated
                unchanged_dict = dialogue_stats.other_initiated
            elif changed_agent == "other":
                changed_dict = dialogue_stats.other_initiated
                unchanged_dict = dialogue_stats.self_initiated
            else:
                raise SyntaxError(
                    f"changed_agent can only be 'self' or 'other'. Found {changed_agent}."
                )

            for end_state_numbers in unchanged_dict.values():
                assert end_state_numbers == 0
            for end_state, end_state_numbers in changed_dict.items():
                if end_state == changed_end_state:
                    assert end_state_numbers == 1
                else:
                    assert end_state_numbers == 0
        else:
            raise SyntaxError(
                "changed_agent and changed_end_state should either both be None, or neither."
            )

    def test_handle_cfp_i(self):
        """Test the _on_cfp method of the fipa handler where proposal_for_query is None."""
        # setup
        mocked_proposal = None
        incoming_message = self.build_incoming_message(
            message_type=FipaMessage,
            performative=FipaMessage.Performative.CFP,
            query=self.cfp_query,
        )

        # before
        self._assert_stat_state(self.dialogue_stats)

        # operation
        with patch.object(
            self.strategy, "get_proposal_for_query", return_value=mocked_proposal,
        ):
            with patch.object(self.logger, "log") as mock_logger:
                self.fipa_handler.handle(incoming_message)

        # after
        self.assert_quantity_in_outbox(1)

        mock_logger.assert_any_call(
            logging.INFO,
            f"received {incoming_message.performative} from {incoming_message.sender[-5:]} (as {self.fipa_dialogues.get_dialogue(incoming_message).role}), message={incoming_message}",
        )

        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_outbox(),
            message_type=FipaMessage,
            performative=FipaMessage.Performative.DECLINE,
            to=incoming_message.sender,
            sender=self.skill.skill_context.agent_address,
            target=incoming_message.message_id,
        )
        assert has_attributes, error_str

        self._assert_stat_state(
            self.dialogue_stats, "other", FipaDialogue.EndState.DECLINED_CFP
        )

    def test_handle_cfp_ii(self):
        """Test the _on_cfp method of the fipa handler where proposal_for_query is NOT None."""
        # setup
        incoming_message = self.build_incoming_message(
            message_type=FipaMessage,
            performative=FipaMessage.Performative.CFP,
            query=self.cfp_query,
        )

        # operation
        with patch.object(
            self.strategy, "get_proposal_for_query", return_value=self.proposal
        ):
            with patch.object(self.strategy, "terms_from_proposal") as mock_terms:
                with patch.object(
                    self.transactions, "add_pending_proposal"
                ) as mock_pending:
                    with patch.object(self.logger, "log") as mock_logger:
                        self.fipa_handler.handle(incoming_message)

        # after
        self.assert_quantity_in_outbox(1)

        mock_logger.assert_any_call(
            logging.INFO,
            f"received {incoming_message.performative} from {incoming_message.sender[-5:]} (as {self.fipa_dialogues.get_dialogue(incoming_message).role}), message={incoming_message}",
        )

        message = self.get_message_from_outbox()
        has_attributes, error_str = self.message_has_attributes(
            actual_message=message,
            message_type=FipaMessage,
            performative=FipaMessage.Performative.PROPOSE,
            to=incoming_message.sender,
            sender=self.skill.skill_context.agent_address,
            target=incoming_message.message_id,
            proposal=self.proposal,
        )
        assert has_attributes, error_str

        mock_terms.assert_called_once()
        mock_pending.assert_called_once()

        mock_logger.assert_any_call(
            logging.INFO,
            f"sending {message.performative} to {message.to[-5:]} (as {self.fipa_dialogues.get_dialogue(message).role}), message={message}",
        )

    def test_handle_propose_i(self):
        """Test the _handle_propose method of the fipa handler where the tx IS profitable."""
        # setup
        fipa_dialogue = self.prepare_skill_dialogue(
            dialogues=self.fipa_dialogues,
            messages=self.list_of_messages_self_initiated[:1],
        )
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=fipa_dialogue,
            performative=FipaMessage.Performative.PROPOSE,
            proposal=self.proposal,
        )

        # operation
        with patch.object(
            self.strategy, "is_profitable_transaction", return_value=True
        ):
            with patch.object(self.transactions, "add_locked_tx") as mock_lock:
                with patch.object(
                    self.transactions, "add_pending_initial_acceptance"
                ) as mock_pending:
                    with patch.object(self.logger, "log") as mock_logger:
                        self.fipa_handler.handle(incoming_message)

        # after
        self.assert_quantity_in_outbox(1)

        mock_logger.assert_any_call(
            logging.INFO,
            f"received {incoming_message.performative} from {incoming_message.sender[-5:]} (as {self.fipa_dialogues.get_dialogue(incoming_message).role}), message={incoming_message}",
        )

        message = self.get_message_from_outbox()
        has_attributes, error_str = self.message_has_attributes(
            actual_message=message,
            message_type=FipaMessage,
            performative=FipaMessage.Performative.ACCEPT,
            to=incoming_message.sender,
            sender=self.skill.skill_context.agent_address,
            target=incoming_message.message_id,
        )
        assert has_attributes, error_str

        mock_lock.assert_called_once()
        mock_pending.assert_called_once()

        mock_logger.assert_any_call(
            logging.INFO,
            f"sending {message.performative} to {message.to[-5:]} (as {self.fipa_dialogues.get_dialogue(message).role}), message={message}",
        )

    def test_handle_propose_ii(self):
        """Test the _handle_propose method of the fipa handler where the tx is NOT profitable."""
        # setup
        fipa_dialogue = self.prepare_skill_dialogue(
            dialogues=self.fipa_dialogues,
            messages=self.list_of_messages_self_initiated[:1],
        )
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=fipa_dialogue,
            performative=FipaMessage.Performative.PROPOSE,
            proposal=self.proposal,
        )

        # before
        self._assert_stat_state(self.dialogue_stats)

        # operation
        with patch.object(
            self.strategy, "is_profitable_transaction", return_value=False
        ):
            with patch.object(self.logger, "log") as mock_logger:
                self.fipa_handler.handle(incoming_message)

        # after
        self.assert_quantity_in_outbox(1)

        mock_logger.assert_any_call(
            logging.INFO,
            f"received {incoming_message.performative} from {incoming_message.sender[-5:]} (as {self.fipa_dialogues.get_dialogue(incoming_message).role}), message={incoming_message}",
        )

        message = self.get_message_from_outbox()
        has_attributes, error_str = self.message_has_attributes(
            actual_message=message,
            message_type=FipaMessage,
            performative=FipaMessage.Performative.DECLINE,
            to=incoming_message.sender,
            sender=self.skill.skill_context.agent_address,
            target=incoming_message.message_id,
        )
        assert has_attributes, error_str

        self._assert_stat_state(
            self.dialogue_stats, "self", FipaDialogue.EndState.DECLINED_PROPOSE
        )

        mock_logger.assert_any_call(
            logging.INFO,
            f"sending {message.performative} to {message.to[-5:]} (as {self.fipa_dialogues.get_dialogue(message).role}), message={message}",
        )

    def test_handle_decline_decline_cfp(self):
        """Test the _handle_decline method of the fipa handler where the end state is decline_cfp."""
        # setup
        fipa_dialogue = self.prepare_skill_dialogue(
            dialogues=self.fipa_dialogues,
            messages=self.list_of_messages_self_initiated[:1],
        )
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=fipa_dialogue, performative=FipaMessage.Performative.DECLINE,
        )

        # before
        self._assert_stat_state(self.dialogue_stats)

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.fipa_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received {incoming_message.performative} from {incoming_message.sender[-5:]} (as {self.fipa_dialogues.get_dialogue(incoming_message).role}), message={incoming_message}",
        )

        self._assert_stat_state(
            self.dialogue_stats, "self", FipaDialogue.EndState.DECLINED_CFP
        )

    def test_handle_decline_decline_propose(self):
        """Test the _handle_decline method of the fipa handler where the end state is decline_propose."""
        # setup
        fipa_dialogue = self.prepare_skill_dialogue(
            dialogues=self.fipa_dialogues,
            messages=self.list_of_messages_other_initiated[:2],
        )
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=fipa_dialogue, performative=FipaMessage.Performative.DECLINE,
        )

        # before
        self._assert_stat_state(self.dialogue_stats)

        # operation
        with patch.object(self.transactions, "pop_pending_proposal") as mock_pending:
            with patch.object(self.logger, "log") as mock_logger:
                self.fipa_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received {incoming_message.performative} from {incoming_message.sender[-5:]} (as {self.fipa_dialogues.get_dialogue(incoming_message).role}), message={incoming_message}",
        )

        self._assert_stat_state(
            self.dialogue_stats, "other", FipaDialogue.EndState.DECLINED_PROPOSE
        )

        mock_pending.assert_called_once()

    def test_handle_decline_decline_accept(self):
        """Test the _handle_decline method of the fipa handler where the end state is decline_accept."""
        # setup
        fipa_dialogue = self.prepare_skill_dialogue(
            dialogues=self.fipa_dialogues,
            messages=self.list_of_messages_self_initiated[:3],
        )
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=fipa_dialogue, performative=FipaMessage.Performative.DECLINE,
        )

        # before
        self._assert_stat_state(self.dialogue_stats)

        # operation
        with patch.object(
            self.transactions, "pop_pending_initial_acceptance"
        ) as mock_pending:
            with patch.object(self.transactions, "pop_locked_tx") as mock_locked:
                with patch.object(self.logger, "log") as mock_logger:
                    self.fipa_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received {incoming_message.performative} from {incoming_message.sender[-5:]} (as {self.fipa_dialogues.get_dialogue(incoming_message).role}), message={incoming_message}",
        )

        self._assert_stat_state(
            self.dialogue_stats, "self", FipaDialogue.EndState.DECLINED_ACCEPT
        )

        mock_pending.assert_called_once()
        mock_locked.assert_called_once()

    def test_handle_accept_i(self):
        """Test the _on_accept method of the fipa handler where the tx IS profitable and strategy's is_contract_tx is True."""
        # setup
        self.strategy._is_contract_tx = True

        fipa_dialogue = cast(
            FipaDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.fipa_dialogues,
                messages=self.list_of_messages_other_initiated[:2],
                is_agent_to_agent_messages=True,
            ),
        )
        fipa_dialogue._terms = self.terms
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=fipa_dialogue, performative=FipaMessage.Performative.ACCEPT,
        )

        # operation
        with patch.object(self.transactions, "pop_pending_proposal") as mock_pending:
            with patch.object(
                self.strategy, "is_profitable_transaction", return_value=True
            ):
                with patch.object(self.transactions, "add_locked_tx") as mock_lock:
                    with patch.object(
                        type(self.strategy),
                        "ledger_id",
                        new_callable=PropertyMock,
                        return_value=self.ledger_id,
                    ):
                        with patch.object(
                            type(self.strategy),
                            "contract_id",
                            new_callable=PropertyMock,
                            return_value=self.contract_id,
                        ):
                            with patch.object(
                                type(self.strategy),
                                "contract_address",
                                new_callable=PropertyMock,
                                return_value=self.contract_address,
                            ):
                                with patch.object(
                                    self.strategy,
                                    "kwargs_from_terms",
                                    return_value=self.kwargs,
                                ):
                                    with patch.object(
                                        self.logger, "log"
                                    ) as mock_logger:
                                        self.fipa_handler.handle(incoming_message)

        # after
        self.assert_quantity_in_outbox(1)

        mock_logger.assert_any_call(
            logging.INFO,
            f"received {incoming_message.performative} from {incoming_message.sender[-5:]} (as {self.fipa_dialogues.get_dialogue(incoming_message).role}), message={incoming_message}",
        )

        mock_pending.assert_called_once()
        mock_lock.assert_called_once()

        message = self.get_message_from_outbox()
        has_attributes, error_str = self.message_has_attributes(
            actual_message=message,
            message_type=ContractApiMessage,
            performative=ContractApiMessage.Performative.GET_RAW_MESSAGE,
            to=LEDGER_API_ADDRESS,
            sender=str(self.skill.skill_context.skill_id),
            ledger_id=self.ledger_id,
            contract_id=self.contract_id,
            contract_address=self.contract_address,
            callable="get_hash_batch",
            kwargs=ContractApiMessage.Kwargs(self.kwargs),
        )
        assert has_attributes, error_str

        assert (
            cast(
                ContractApiDialogue, self.contract_api_dialogues.get_dialogue(message)
            ).associated_fipa_dialogue
            == fipa_dialogue
        )

        mock_logger.assert_any_call(
            logging.INFO,
            f"requesting batch transaction hash, sending {message.performative} to {self.contract_id}, message={message}",
        )

    def test_handle_accept_ii(self):
        """Test the _on_accept method of the fipa handler where the tx IS profitable and strategy's is_contract_tx is False."""
        # setup
        self.strategy._is_contract_tx = False

        fipa_dialogue = cast(
            FipaDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.fipa_dialogues,
                messages=self.list_of_messages_other_initiated[:2],
            ),
        )
        fipa_dialogue._terms = self.terms
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=fipa_dialogue, performative=FipaMessage.Performative.ACCEPT,
        )

        # operation
        with patch.object(
            self.transactions, "pop_pending_proposal", return_value=self.terms
        ) as mock_pending:
            with patch.object(
                self.strategy, "is_profitable_transaction", return_value=True
            ):
                with patch.object(self.logger, "log") as mock_logger:
                    self.fipa_handler.handle(incoming_message)

        # after
        self.assert_quantity_in_decision_making_queue(1)

        mock_logger.assert_any_call(
            logging.INFO,
            f"received {incoming_message.performative} from {incoming_message.sender[-5:]} (as {self.fipa_dialogues.get_dialogue(incoming_message).role}), message={incoming_message}",
        )

        mock_pending.assert_called_once()

        message = self.get_message_from_decision_maker_inbox()
        has_attributes, error_str = self.message_has_attributes(
            actual_message=message,
            message_type=SigningMessage,
            performative=SigningMessage.Performative.SIGN_MESSAGE,
            to=self.skill.skill_context.decision_maker_address,
            sender=str(self.skill.skill_context.skill_id),
            terms=self.terms,
            raw_message=self.raw_message,
        )
        assert has_attributes, error_str

        assert (
            cast(
                SigningDialogue, self.signing_dialogues.get_dialogue(message)
            ).associated_fipa_dialogue
            == fipa_dialogue
        )

        mock_logger.assert_any_call(
            logging.INFO,
            f"requesting signature, sending {message.performative} to decision_maker, message={message}",
        )

    def test_handle_accept_iii(self):
        """Test the _on_accept method of the fipa handler where the tx is NOT profitable."""
        # setup
        fipa_dialogue = cast(
            FipaDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.fipa_dialogues,
                messages=self.list_of_messages_other_initiated[:2],
            ),
        )
        fipa_dialogue._terms = self.terms
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=fipa_dialogue, performative=FipaMessage.Performative.ACCEPT,
        )

        # before
        self._assert_stat_state(self.dialogue_stats)

        # operation
        with patch.object(
            self.transactions, "pop_pending_proposal", return_value=self.terms
        ) as mock_pending:
            with patch.object(
                self.strategy, "is_profitable_transaction", return_value=False
            ):
                with patch.object(self.logger, "log") as mock_logger:
                    self.fipa_handler.handle(incoming_message)

        # after
        self.assert_quantity_in_outbox(1)

        mock_logger.assert_any_call(
            logging.INFO,
            f"received {incoming_message.performative} from {incoming_message.sender[-5:]} (as {self.fipa_dialogues.get_dialogue(incoming_message).role}), message={incoming_message}",
        )

        mock_pending.assert_called_once()

        message = self.get_message_from_outbox()
        has_attributes, error_str = self.message_has_attributes(
            actual_message=message,
            message_type=FipaMessage,
            performative=FipaMessage.Performative.DECLINE,
            to=incoming_message.sender,
            sender=self.skill.skill_context.agent_address,
            target=incoming_message.message_id,
        )
        assert has_attributes, error_str

        self._assert_stat_state(
            self.dialogue_stats, "other", FipaDialogue.EndState.DECLINED_ACCEPT
        )

        mock_logger.assert_any_call(
            logging.INFO,
            f"sending {message.performative} to {message.to[-5:]} (as {self.fipa_dialogues.get_dialogue(message).role}), message={message}",
        )

    def test_handle_match_accept_i(self):
        """Test the _handle_match_accept method of the fipa handler where is_contract_tx is True and counterparty signature is None."""
        # setup
        self.strategy._is_contract_tx = True

        fipa_dialogue = self.prepare_skill_dialogue(
            dialogues=self.fipa_dialogues,
            messages=self.list_of_messages_self_initiated[:3],
        )
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=fipa_dialogue,
            performative=FipaMessage.Performative.MATCH_ACCEPT_W_INFORM,
            info={"signature": None},
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.fipa_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received {incoming_message.performative} from {incoming_message.sender[-5:]} (as {self.fipa_dialogues.get_dialogue(incoming_message).role}), message={incoming_message}",
        )

        mock_logger.assert_any_call(
            logging.INFO,
            f"{incoming_message.performative} did not contain counterparty signature!",
        )

    def test_handle_match_accept_ii(self):
        """Test the _handle_match_accept method of the fipa handler where is_contract_tx is True and counterparty signature is NOT None."""
        # setup
        self.strategy._is_contract_tx = True

        fipa_dialogue = cast(
            FipaDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.fipa_dialogues,
                messages=self.list_of_messages_self_initiated[:3],
            ),
        )
        fipa_dialogue._terms = self.terms
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=fipa_dialogue,
            performative=FipaMessage.Performative.MATCH_ACCEPT_W_INFORM,
            info={"signature": self.counterparty_signature},
        )

        # operation
        with patch.object(
            type(self.strategy),
            "ledger_id",
            new_callable=PropertyMock,
            return_value=self.ledger_id,
        ):
            with patch.object(
                type(self.strategy),
                "contract_id",
                new_callable=PropertyMock,
                return_value=self.contract_id,
            ):
                with patch.object(
                    type(self.strategy),
                    "contract_address",
                    new_callable=PropertyMock,
                    return_value=self.contract_address,
                ):
                    with patch.object(
                        self.strategy, "kwargs_from_terms", return_value=self.kwargs
                    ):
                        with patch.object(self.logger, "log") as mock_logger:
                            self.fipa_handler.handle(incoming_message)

        # after
        self.assert_quantity_in_outbox(1)

        mock_logger.assert_any_call(
            logging.INFO,
            f"received {incoming_message.performative} from {incoming_message.sender[-5:]} (as {self.fipa_dialogues.get_dialogue(incoming_message).role}), message={incoming_message}",
        )

        assert fipa_dialogue.counterparty_signature == self.counterparty_signature

        message = self.get_message_from_outbox()
        has_attributes, error_str = self.message_has_attributes(
            actual_message=message,
            message_type=ContractApiMessage,
            performative=ContractApiMessage.Performative.GET_RAW_TRANSACTION,
            to=LEDGER_API_ADDRESS,
            sender=str(self.skill.skill_context.skill_id),
            ledger_id=self.ledger_id,
            contract_id=self.contract_id,
            contract_address=self.contract_address,
            callable="get_atomic_swap_batch_transaction",
            kwargs=ContractApiMessage.Kwargs(self.kwargs),
        )
        assert has_attributes, error_str

        assert (
            cast(
                ContractApiDialogue, self.contract_api_dialogues.get_dialogue(message)
            ).associated_fipa_dialogue
            == fipa_dialogue
        )

        mock_logger.assert_any_call(
            logging.INFO,
            f"requesting batch atomic swap transaction, sending {message.performative} to {self.contract_id}, message={message}",
        )

    def test_handle_match_accept_iii(self):
        """Test the _handle_match_accept method of the fipa handler where is_contract_tx is False and counterparty signature is NOT None."""
        # setup
        self.strategy._is_contract_tx = False

        fipa_dialogue = cast(
            FipaDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.fipa_dialogues,
                messages=self.list_of_messages_self_initiated[:3],
            ),
        )
        fipa_dialogue._terms = self.terms
        incoming_message = self.build_incoming_message_for_skill_dialogue(
            dialogue=fipa_dialogue,
            performative=FipaMessage.Performative.MATCH_ACCEPT_W_INFORM,
            info={"signature": self.counterparty_signature},
        )

        # operation
        with patch.object(
            self.transactions, "pop_pending_initial_acceptance", return_value=self.terms
        ) as mock_pending:
            with patch.object(self.logger, "log") as mock_logger:
                self.fipa_handler.handle(incoming_message)

        # after
        self.assert_quantity_in_decision_making_queue(1)

        mock_logger.assert_any_call(
            logging.INFO,
            f"received {incoming_message.performative} from {incoming_message.sender[-5:]} (as {self.fipa_dialogues.get_dialogue(incoming_message).role}), message={incoming_message}",
        )

        mock_pending.assert_called_once()

        assert fipa_dialogue.counterparty_signature == self.counterparty_signature

        message = self.get_message_from_decision_maker_inbox()
        has_attributes, error_str = self.message_has_attributes(
            actual_message=message,
            message_type=SigningMessage,
            performative=SigningMessage.Performative.SIGN_MESSAGE,
            to=self.skill.skill_context.decision_maker_address,
            sender=str(self.skill.skill_context.skill_id),
            terms=self.terms,
            raw_message=self.raw_message,
        )
        assert has_attributes, error_str

        assert (
            cast(
                SigningDialogue, self.signing_dialogues.get_dialogue(message)
            ).associated_fipa_dialogue
            == fipa_dialogue
        )

        mock_logger.assert_any_call(
            logging.INFO,
            f"requesting signature, sending {message.performative} to decision_maker, message={message}",
        )

    def test_teardown(self):
        """Test the teardown method of the fipa handler."""
        assert self.fipa_handler.teardown() is None
        self.assert_quantity_in_outbox(0)


class TestSigningHandler(BaseSkillTestCase):
    """Test signing handler of tac negotiation."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "tac_negotiation")
    is_agent_to_agent_messages = False

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.signing_handler = cast(
            SigningHandler, cls._skill.skill_context.handlers.signing
        )
        cls.strategy = cast(Strategy, cls._skill.skill_context.strategy)
        cls.logger = cls.signing_handler.context.logger

        cls.signing_dialogues = cast(
            SigningDialogues, cls._skill.skill_context.signing_dialogues
        )
        cls.fipa_dialogues = cast(
            FipaDialogues, cls._skill.skill_context.fipa_dialogues
        )
        cls.ledger_api_dialogues = cast(
            LedgerApiDialogues, cls._skill.skill_context.ledger_api_dialogues
        )

        cls.ledger_id = "some_ledger_id"
        cls.nonce = "some_nonce"
        cls.body = {"some_key": "some_value"}
        cls.body_bytes = b"some_body"
        cls.body_str = "some_body"
        cls.counterparty_signature = "some_counterparty_signature"
        cls.terms = Terms(
            "some_ledger_id",
            cls._skill.skill_context.agent_address,
            "counterprty",
            {"currency_id": 50},
            {"good_id": -10},
            "some_nonce",
        )
        cls.list_of_signing_msg_messages = (
            DialogueMessage(
                SigningMessage.Performative.SIGN_MESSAGE,
                {
                    "terms": cls.terms,
                    "raw_message": SigningMessage.RawMessage(
                        cls.ledger_id, cls.body_bytes
                    ),
                },
            ),
        )
        cls.list_of_signing_tx_messages = (
            DialogueMessage(
                SigningMessage.Performative.SIGN_TRANSACTION,
                {
                    "terms": cls.terms,
                    "raw_transaction": SigningMessage.RawTransaction(
                        cls.ledger_id, cls.body
                    ),
                },
            ),
        )

        cls.cfp_query = Query(
            [Constraint("some_attribute", ConstraintType("==", "some_service"))],
            DataModel(
                SUPPLY_DATAMODEL_NAME,
                [
                    Attribute(
                        "some_attribute", str, False, "Some attribute descriptions."
                    )
                ],
            ),
        )
        cls.proposal = Description(
            {
                "ledger_id": cls.ledger_id,
                "price": 100,
                "currency_id": "1",
                "fee": 1,
                "nonce": cls.nonce,
            }
        )
        cls.list_of_other_initiated_fipa_messages = (
            DialogueMessage(
                FipaMessage.Performative.CFP, {"query": cls.cfp_query}, True
            ),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE, {"proposal": cls.proposal}
            ),
            DialogueMessage(FipaMessage.Performative.ACCEPT),
            DialogueMessage(
                FipaMessage.Performative.MATCH_ACCEPT_W_INFORM,
                {"info": {"address": "some_term_sender_address"}},
            ),
            DialogueMessage(
                FipaMessage.Performative.INFORM,
                {"info": {"transaction_digest": "some_transaction_digest_body"}},
            ),
        )
        cls.list_of_self_initiated_fipa_messages = (
            DialogueMessage(FipaMessage.Performative.CFP, {"query": cls.cfp_query}),
            DialogueMessage(
                FipaMessage.Performative.PROPOSE, {"proposal": cls.proposal}
            ),
            DialogueMessage(FipaMessage.Performative.ACCEPT),
            DialogueMessage(
                FipaMessage.Performative.MATCH_ACCEPT_W_INFORM,
                {"info": {"address": "some_term_sender_address"}},
            ),
            DialogueMessage(
                FipaMessage.Performative.INFORM,
                {"info": {"transaction_digest": "some_transaction_digest_body"}},
            ),
        )

    def test_setup(self):
        """Test the setup method of the signing handler."""
        assert self.signing_handler.setup() is None
        self.assert_quantity_in_outbox(0)

    def test_handle_unidentified_dialogue(self):
        """Test the _handle_unidentified_dialogue method of the signing handler."""
        # setup
        incorrect_dialogue_reference = ("", "")
        incoming_message = self.build_incoming_message(
            message_type=SigningMessage,
            dialogue_reference=incorrect_dialogue_reference,
            performative=SigningMessage.Performative.ERROR,
            error_code=SigningMessage.ErrorCode.UNSUCCESSFUL_MESSAGE_SIGNING,
            to=str(self.skill.skill_context.skill_id),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.signing_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received invalid signing message={incoming_message}, unidentified dialogue.",
        )

    def test_handle_signed_message_i(self):
        """Test the _handle_signed_message method of the signing handler where last fipa message is ACCEPT."""
        # setup
        fipa_dialogue = cast(
            FipaDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.fipa_dialogues,
                messages=self.list_of_other_initiated_fipa_messages[:3],
                is_agent_to_agent_messages=True,
            ),
        )

        signing_dialogue = cast(
            SigningDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.signing_dialogues,
                messages=self.list_of_signing_msg_messages[:1],
                counterparty=self.skill.skill_context.decision_maker_address,
            ),
        )
        signing_dialogue.associated_fipa_dialogue = fipa_dialogue

        incoming_message = cast(
            SigningMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=signing_dialogue,
                performative=SigningMessage.Performative.SIGNED_MESSAGE,
                signed_message=SigningMessage.SignedMessage(
                    self.ledger_id, self.body_str
                ),
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.signing_handler.handle(incoming_message)

        # after
        self.assert_quantity_in_outbox(1)

        mock_logger.assert_any_call(
            logging.INFO,
            f"received {incoming_message.performative} from decision_maker, message={incoming_message}",
        )

        message = self.get_message_from_outbox()
        has_attributes, error_str = self.message_has_attributes(
            actual_message=message,
            message_type=FipaMessage,
            performative=FipaMessage.Performative.MATCH_ACCEPT_W_INFORM,
            to=fipa_dialogue.dialogue_label.dialogue_opponent_addr,
            # (line below) match-accept is already added to fipa_dialogue, hence "-1"
            target=fipa_dialogue.last_incoming_message.message_id,
            sender=self.skill.skill_context.agent_address,
            info={"signature": incoming_message.signed_message.body},
        )
        assert has_attributes, error_str

        mock_logger.assert_any_call(
            logging.INFO,
            f"sending {message.performative.value} to {message.to[-5:]} (as {fipa_dialogue.role}), message={message}.",
        )

    def test_handle_signed_message_ii(self):
        """Test the _handle_signed_message method of the signing handler where last fipa message is MATCH_ACCEPT."""
        # setup
        mocked_tx = {
            "terms": self.terms,
            "sender_signature": self.body_str,
            "counterparty_signature": self.counterparty_signature,
        }

        fipa_dialogue = cast(
            FipaDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.fipa_dialogues,
                messages=self.list_of_self_initiated_fipa_messages[:4],
                is_agent_to_agent_messages=True,
            ),
        )
        fipa_dialogue.counterparty_signature = self.counterparty_signature
        fipa_dialogue.terms = self.terms

        signing_dialogue = cast(
            SigningDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.signing_dialogues,
                messages=self.list_of_signing_msg_messages[:1],
                counterparty=self.skill.skill_context.decision_maker_address,
            ),
        )
        signing_dialogue.associated_fipa_dialogue = fipa_dialogue

        incoming_message = cast(
            SigningMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=signing_dialogue,
                performative=SigningMessage.Performative.SIGNED_MESSAGE,
                signed_message=SigningMessage.SignedMessage(
                    self.ledger_id, self.body_str
                ),
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.signing_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received {incoming_message.performative} from decision_maker, message={incoming_message}",
        )

        assert (
            self.skill.skill_context.shared_state["transactions"][
                fipa_dialogue.terms.sender_hash
            ]
            == mocked_tx
        )

        mock_logger.assert_any_call(
            logging.INFO, f"sending transaction to controller, tx={mocked_tx}."
        )

    def test_handle_signed_message_iii(self):
        """Test the _handle_signed_message method of the signing handler where last fipa message is neither ACCEPT nor MATCH_ACCEPT."""
        # setup
        fipa_dialogue = cast(
            FipaDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.fipa_dialogues,
                messages=self.list_of_self_initiated_fipa_messages[:3],
                is_agent_to_agent_messages=True,
            ),
        )

        signing_dialogue = cast(
            SigningDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.signing_dialogues,
                messages=self.list_of_signing_msg_messages[:1],
                counterparty=self.skill.skill_context.decision_maker_address,
            ),
        )
        signing_dialogue.associated_fipa_dialogue = fipa_dialogue

        incoming_message = cast(
            SigningMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=signing_dialogue,
                performative=SigningMessage.Performative.SIGNED_MESSAGE,
                signed_message=SigningMessage.SignedMessage(
                    self.ledger_id, self.body_str
                ),
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            with pytest.raises(
                AEAEnforceError,
                match="last message should be of performative accept or match accept.",
            ):
                self.signing_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received {incoming_message.performative} from decision_maker, message={incoming_message}",
        )

    def test_handle_signed_message_iv(self):
        """Test the _handle_signed_message method of the signing handler where last fipa message is None."""
        # setup
        fipa_dialogue = cast(
            FipaDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.fipa_dialogues,
                messages=self.list_of_self_initiated_fipa_messages[:1],
                is_agent_to_agent_messages=True,
            ),
        )
        fipa_dialogue._incoming_messages = []
        fipa_dialogue._outgoing_messages = []

        signing_dialogue = cast(
            SigningDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.signing_dialogues,
                messages=self.list_of_signing_msg_messages[:1],
                counterparty=self.skill.skill_context.decision_maker_address,
            ),
        )
        signing_dialogue.associated_fipa_dialogue = fipa_dialogue

        incoming_message = cast(
            SigningMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=signing_dialogue,
                performative=SigningMessage.Performative.SIGNED_MESSAGE,
                signed_message=SigningMessage.SignedMessage(
                    self.ledger_id, self.body_str
                ),
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            with pytest.raises(AEAEnforceError, match="last message not recovered."):
                self.signing_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received {incoming_message.performative} from decision_maker, message={incoming_message}",
        )

    def test_handle_signed_transaction_i(self):
        """Test the _handle_signed_transaction method of the signing handler where is_contract_tx is True and last fipa message is ACCEPT."""
        # setup
        self.strategy._is_contract_tx = True

        fipa_dialogue = cast(
            FipaDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.fipa_dialogues,
                messages=self.list_of_other_initiated_fipa_messages[:3],
                is_agent_to_agent_messages=True,
            ),
        )

        signing_dialogue = cast(
            SigningDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.signing_dialogues,
                messages=self.list_of_signing_tx_messages[:1],
                counterparty=self.skill.skill_context.decision_maker_address,
            ),
        )
        signing_dialogue.associated_fipa_dialogue = fipa_dialogue

        incoming_message = cast(
            SigningMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=signing_dialogue,
                performative=SigningMessage.Performative.SIGNED_TRANSACTION,
                signed_transaction=SigningMessage.SignedTransaction(
                    self.ledger_id, self.body
                ),
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.signing_handler.handle(incoming_message)

        # after
        self.assert_quantity_in_outbox(1)

        mock_logger.assert_any_call(
            logging.INFO,
            f"received {incoming_message.performative} from decision_maker, message={incoming_message}",
        )

        message = self.get_message_from_outbox()
        has_attributes, error_str = self.message_has_attributes(
            actual_message=message,
            message_type=FipaMessage,
            performative=FipaMessage.Performative.MATCH_ACCEPT_W_INFORM,
            to=fipa_dialogue.dialogue_label.dialogue_opponent_addr,
            # (line below) match-accept is already added to fipa_dialogue, hence "-1"
            sender=self.skill.skill_context.agent_address,
            info={"tx_signature": incoming_message.signed_transaction},
        )
        assert has_attributes, error_str

        mock_logger.assert_any_call(
            logging.INFO,
            f"sending {message.performative.value} to {message.to[-5:]} (as {fipa_dialogue.role}), message={message}.",
        )

    def test_handle_signed_transaction_ii(self):
        """Test the _handle_signed_transaction method of the signing handler where is_contract_tx is True and last fipa message is MATCH_ACCEPT."""
        # setup
        self.strategy._is_contract_tx = True

        fipa_dialogue = cast(
            FipaDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.fipa_dialogues,
                messages=self.list_of_self_initiated_fipa_messages[:4],
                is_agent_to_agent_messages=True,
            ),
        )

        signing_dialogue = cast(
            SigningDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.signing_dialogues,
                messages=self.list_of_signing_tx_messages[:1],
                counterparty=self.skill.skill_context.decision_maker_address,
            ),
        )
        signing_dialogue.associated_fipa_dialogue = fipa_dialogue

        incoming_message = cast(
            SigningMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=signing_dialogue,
                performative=SigningMessage.Performative.SIGNED_TRANSACTION,
                signed_transaction=SigningMessage.SignedTransaction(
                    self.ledger_id, self.body
                ),
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.signing_handler.handle(incoming_message)

        # after
        self.assert_quantity_in_outbox(1)

        mock_logger.assert_any_call(
            logging.INFO,
            f"received {incoming_message.performative} from decision_maker, message={incoming_message}",
        )

        message = self.get_message_from_outbox()
        has_attributes, error_str = self.message_has_attributes(
            actual_message=message,
            message_type=LedgerApiMessage,
            performative=LedgerApiMessage.Performative.SEND_SIGNED_TRANSACTION,
            to=LEDGER_API_ADDRESS,
            sender=str(self.skill.skill_context.skill_id),
            signed_transaction=incoming_message.signed_transaction,
        )
        assert has_attributes, error_str

        assert (
            cast(
                LedgerApiDialogue, self.ledger_api_dialogues.get_dialogue(message)
            ).associated_signing_dialogue
            == signing_dialogue
        )

        mock_logger.assert_any_call(
            logging.INFO,
            f"sending {message.performative} to ledger {self.strategy.ledger_id}, message={message}",
        )

    def test_handle_signed_transaction_iii(self):
        """Test the _handle_signed_transaction method of the signing handler where is_contract_tx is True and last fipa message is neither ACCEPT nor MATCH_ACCEPT."""
        # setup
        self.strategy._is_contract_tx = True

        fipa_dialogue = cast(
            FipaDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.fipa_dialogues,
                messages=self.list_of_self_initiated_fipa_messages[:3],
                is_agent_to_agent_messages=True,
            ),
        )

        signing_dialogue = cast(
            SigningDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.signing_dialogues,
                messages=self.list_of_signing_tx_messages[:1],
                counterparty=self.skill.skill_context.decision_maker_address,
            ),
        )
        signing_dialogue.associated_fipa_dialogue = fipa_dialogue

        incoming_message = cast(
            SigningMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=signing_dialogue,
                performative=SigningMessage.Performative.SIGNED_TRANSACTION,
                signed_transaction=SigningMessage.SignedTransaction(
                    self.ledger_id, self.body
                ),
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            with pytest.raises(
                AEAEnforceError,
                match="last message should be of performative accept or match accept.",
            ):
                self.signing_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received {incoming_message.performative} from decision_maker, message={incoming_message}",
        )

    def test_handle_signed_transaction_iv(self):
        """Test the _handle_signed_transaction method of the signing handler where is_contract_tx is True and last incoming fipa message is None."""
        # setup
        self.strategy._is_contract_tx = True

        fipa_dialogue = cast(
            FipaDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.fipa_dialogues,
                messages=self.list_of_self_initiated_fipa_messages[:3],
                is_agent_to_agent_messages=True,
            ),
        )
        fipa_dialogue._incoming_messages = []
        fipa_dialogue._outgoing_messages = []

        signing_dialogue = cast(
            SigningDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.signing_dialogues,
                messages=self.list_of_signing_tx_messages[:1],
                counterparty=self.skill.skill_context.decision_maker_address,
            ),
        )
        signing_dialogue.associated_fipa_dialogue = fipa_dialogue

        incoming_message = cast(
            SigningMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=signing_dialogue,
                performative=SigningMessage.Performative.SIGNED_TRANSACTION,
                signed_transaction=SigningMessage.SignedTransaction(
                    self.ledger_id, self.body
                ),
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            with pytest.raises(AEAEnforceError, match="last message not recovered."):
                self.signing_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received {incoming_message.performative} from decision_maker, message={incoming_message}",
        )

    def test_handle_signed_transaction_v(self):
        """Test the _handle_signed_transaction method of the signing handler where is_contract_tx is False."""
        # setup
        self.strategy._is_contract_tx = False

        signing_dialogue = cast(
            SigningDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.signing_dialogues,
                messages=self.list_of_signing_tx_messages[:1],
                counterparty=self.skill.skill_context.decision_maker_address,
            ),
        )

        incoming_message = cast(
            SigningMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=signing_dialogue,
                performative=SigningMessage.Performative.SIGNED_TRANSACTION,
                signed_transaction=SigningMessage.SignedTransaction(
                    self.ledger_id, self.body
                ),
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.signing_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received {incoming_message.performative} from decision_maker, message={incoming_message}",
        )

        mock_logger.assert_any_call(
            logging.WARNING, "signed transaction handler only for contract case."
        )

    def test_handle_error(self):
        """Test the _handle_error method of the signing handler."""
        # setup
        signing_counterparty = self.skill.skill_context.decision_maker_address
        signing_dialogue = self.prepare_skill_dialogue(
            dialogues=self.signing_dialogues,
            messages=self.list_of_signing_tx_messages[:1],
            counterparty=signing_counterparty,
        )
        incoming_message = cast(
            SigningMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=signing_dialogue,
                performative=SigningMessage.Performative.ERROR,
                error_code=SigningMessage.ErrorCode.UNSUCCESSFUL_TRANSACTION_SIGNING,
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.signing_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received {incoming_message.performative} from decision_maker, message={incoming_message}",
        )

        mock_logger.assert_any_call(
            logging.INFO,
            f"transaction signing was not successful. Error_code={incoming_message.error_code} in dialogue={signing_dialogue}",
        )

    def test_handle_invalid(self):
        """Test the _handle_invalid method of the signing handler."""
        # setup
        invalid_performative = SigningMessage.Performative.SIGN_TRANSACTION
        incoming_message = self.build_incoming_message(
            message_type=SigningMessage,
            dialogue_reference=("1", ""),
            performative=invalid_performative,
            terms=self.terms,
            raw_transaction=SigningMessage.RawTransaction(
                "some_ledger_id", {"some_key": "some_value"}
            ),
            to=str(self.skill.skill_context.skill_id),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.signing_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received {incoming_message.performative} from decision_maker, message={incoming_message}",
        )

        mock_logger.assert_any_call(
            logging.WARNING,
            f"cannot handle signing message of performative={invalid_performative} in dialogue={self.signing_dialogues.get_dialogue(incoming_message)}.",
        )

    def test_teardown(self):
        """Test the teardown method of the signing handler."""
        assert self.signing_handler.teardown() is None
        self.assert_quantity_in_outbox(0)


class TestLedgerApiHandler(BaseSkillTestCase):
    """Test ledger_api handler of tac negotiation."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "tac_negotiation")
    is_agent_to_agent_messages = False

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.ledger_api_handler = cast(
            LedgerApiHandler, cls._skill.skill_context.handlers.ledger_api
        )
        cls.logger = cls.ledger_api_handler.context.logger

        cls.ledger_api_dialogues = cast(
            LedgerApiDialogues, cls._skill.skill_context.ledger_api_dialogues
        )

        cls.ledger_id = "some_ledger_id"
        cls.contract_id = "some_contract_id"
        cls.callable = "some_callable"
        cls.kwargs = Kwargs({"some_key": "some_value"})

        cls.body = {"some_key": "some_value"}
        cls.body_str = "some_body"
        cls.contract_address = "some_contract_address"

        cls.raw_transaction = RawTransaction(cls.ledger_id, cls.body)
        cls.signed_transaction = SignedTransaction(cls.ledger_id, cls.body)
        cls.transaction_digest = TransactionDigest(cls.ledger_id, cls.body_str)
        cls.receipt = {"contractAddress": cls.contract_address}
        cls.transaction_receipt = TransactionReceipt(
            cls.ledger_id, cls.receipt, {"transaction_key": "transaction_value"}
        )
        cls.address = "some_address"

        cls.terms = Terms(
            cls.ledger_id,
            cls._skill.skill_context.agent_address,
            "counterparty",
            {"currency_id": 50},
            {"good_id": -10},
            "some_nonce",
        )

        cls.list_of_ledger_api_messages = (
            DialogueMessage(
                LedgerApiMessage.Performative.GET_RAW_TRANSACTION, {"terms": cls.terms}
            ),
            DialogueMessage(
                LedgerApiMessage.Performative.RAW_TRANSACTION,
                {"raw_transaction": cls.raw_transaction},
            ),
            DialogueMessage(
                LedgerApiMessage.Performative.SEND_SIGNED_TRANSACTION,
                {"signed_transaction": cls.signed_transaction},
            ),
            DialogueMessage(
                LedgerApiMessage.Performative.TRANSACTION_DIGEST,
                {"transaction_digest": cls.transaction_digest},
            ),
            DialogueMessage(
                LedgerApiMessage.Performative.GET_TRANSACTION_RECEIPT,
                {"transaction_digest": cls.transaction_digest},
            ),
        )

    def test_setup(self):
        """Test the setup method of the ledger_api handler."""
        assert self.ledger_api_handler.setup() is None
        self.assert_quantity_in_outbox(0)

    def test_handle_unidentified_dialogue(self):
        """Test the _handle_unidentified_dialogue method of the ledger_api handler."""
        # setup
        incorrect_dialogue_reference = ("", "")
        incoming_message = self.build_incoming_message(
            message_type=LedgerApiMessage,
            dialogue_reference=incorrect_dialogue_reference,
            performative=LedgerApiMessage.Performative.BALANCE,
            ledger_id="some_ledger_id",
            balance=10,
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.ledger_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received invalid ledger_api message={incoming_message}, unidentified dialogue.",
        )

    def test_handle_balance(self):
        """Test the _handle_balance method of the ledger_api handler."""
        # setup
        balance = 10
        ledger_api_dialogue = cast(
            LedgerApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.ledger_api_dialogues,
                messages=(
                    DialogueMessage(
                        LedgerApiMessage.Performative.GET_BALANCE,
                        {"ledger_id": self.ledger_id, "address": self.address},
                    ),
                ),
                counterparty=LEDGER_API_ADDRESS,
            ),
        )
        incoming_message = cast(
            LedgerApiMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=ledger_api_dialogue,
                performative=LedgerApiMessage.Performative.BALANCE,
                ledger_id=self.ledger_id,
                balance=balance,
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.ledger_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"starting balance on {self.ledger_id} ledger={incoming_message.balance}.",
        )

    def test_handle_transaction_digest(self):
        """Test the _handle_transaction_digest method of the ledger_api handler."""
        # setup
        ledger_api_dialogue = cast(
            LedgerApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.ledger_api_dialogues,
                messages=self.list_of_ledger_api_messages[:3],
                counterparty=LEDGER_API_ADDRESS,
            ),
        )
        incoming_message = cast(
            LedgerApiMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=ledger_api_dialogue,
                performative=LedgerApiMessage.Performative.TRANSACTION_DIGEST,
                transaction_digest=self.transaction_digest,
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.ledger_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"transaction was successfully submitted. Transaction digest={incoming_message.transaction_digest}",
        )

        self.assert_quantity_in_outbox(1)
        has_attributes, error_str = self.message_has_attributes(
            actual_message=self.get_message_from_outbox(),
            message_type=LedgerApiMessage,
            performative=LedgerApiMessage.Performative.GET_TRANSACTION_RECEIPT,
            target=incoming_message.message_id,
            to=LEDGER_API_ADDRESS,
            sender=str(self.skill.skill_context.skill_id),
            transaction_digest=incoming_message.transaction_digest,
        )
        assert has_attributes, error_str

        mock_logger.assert_any_call(logging.INFO, "requesting transaction receipt.")

    def test_handle_transaction_receipt_failed(self):
        """Test the _handle_transaction_receipt method of the ledger_api handler where the transaction is NOT settled."""
        # setup
        ledger_api_dialogue = cast(
            LedgerApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.ledger_api_dialogues,
                messages=self.list_of_ledger_api_messages[:5],
                counterparty=LEDGER_API_ADDRESS,
            ),
        )
        incoming_message = cast(
            LedgerApiMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=ledger_api_dialogue,
                performative=LedgerApiMessage.Performative.TRANSACTION_RECEIPT,
                transaction_receipt=self.transaction_receipt,
            ),
        )

        # operation
        with patch.object(LedgerApis, "is_transaction_settled", return_value=False):
            with patch.object(self.logger, "log") as mock_logger:
                self.ledger_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.ERROR,
            f"transaction failed. Transaction receipt={incoming_message.transaction_receipt}",
        )

    def test_handle_transaction_receipt_succeeds(self):
        """Test the _handle_transaction_receipt method of the ledger_api handler where the transaction is NOT settled."""
        # setup
        ledger_api_dialogue = cast(
            LedgerApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.ledger_api_dialogues,
                messages=self.list_of_ledger_api_messages[:5],
                counterparty=LEDGER_API_ADDRESS,
            ),
        )
        incoming_message = cast(
            LedgerApiMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=ledger_api_dialogue,
                performative=LedgerApiMessage.Performative.TRANSACTION_RECEIPT,
                transaction_receipt=self.transaction_receipt,
            ),
        )

        # operation
        with patch.object(LedgerApis, "is_transaction_settled", return_value=True):
            with patch.object(self.logger, "log") as mock_logger:
                self.ledger_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"transaction was successfully settled. Transaction receipt={incoming_message.transaction_receipt}",
        )

    def test_handle_error(self):
        """Test the _handle_error method of the ledger_api handler."""
        # setup
        ledger_api_dialogue = self.prepare_skill_dialogue(
            dialogues=self.ledger_api_dialogues,
            messages=self.list_of_ledger_api_messages[:1],
        )
        incoming_message = cast(
            LedgerApiMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=ledger_api_dialogue,
                performative=LedgerApiMessage.Performative.ERROR,
                code=1,
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.ledger_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received ledger_api error message={incoming_message} in dialogue={ledger_api_dialogue}.",
        )

    def test_handle_invalid(self):
        """Test the _handle_invalid method of the ledger_api handler."""
        # setup
        invalid_performative = LedgerApiMessage.Performative.GET_BALANCE
        incoming_message = self.build_incoming_message(
            message_type=LedgerApiMessage,
            dialogue_reference=("1", ""),
            performative=invalid_performative,
            ledger_id="some_ledger_id",
            address=self.address,
            to=str(self.skill.skill_context.skill_id),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.ledger_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.WARNING,
            f"cannot handle ledger_api message of performative={invalid_performative} in dialogue={self.ledger_api_dialogues.get_dialogue(incoming_message)}.",
        )

    def test_teardown(self):
        """Test the teardown method of the ledger_api handler."""
        assert self.ledger_api_handler.teardown() is None
        self.assert_quantity_in_outbox(0)


class TestOefSearchHandler(BaseSkillTestCase):
    """Test oef search handler of tac negotiation."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "tac_negotiation")
    is_agent_to_agent_messages = False

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.oef_search_handler = cast(
            OefSearchHandler, cls._skill.skill_context.handlers.oef
        )
        cls.strategy = cast(Strategy, cls._skill.skill_context.strategy)
        cls.logger = cls._skill.skill_context.logger

        cls.oef_dialogues = cast(
            OefSearchDialogues, cls._skill.skill_context.oef_search_dialogues
        )
        cls.fipa_dialogues = cast(
            FipaDialogues, cls._skill.skill_context.fipa_dialogues
        )

        cls.controller_address = "some_controller_address"
        cls.self_address = cls._skill.skill_context.agent_address
        cls.found_agent_address_1 = "some_agent_address_1"
        cls.found_agent_address_2 = "some_agent_address_2"
        cls.found_agent_address_3 = "some_agent_address_3"
        cls.found_agents = [
            cls.self_address,
            cls.found_agent_address_1,
            cls.found_agent_address_2,
            cls.found_agent_address_3,
        ]
        cls.found_agents_less_self = [
            cls.found_agent_address_1,
            cls.found_agent_address_2,
            cls.found_agent_address_3,
        ]
        cls.cfp_query = Query(
            [Constraint("some_attribute", ConstraintType("==", "some_service"))],
            DataModel(
                SUPPLY_DATAMODEL_NAME,
                [
                    Attribute(
                        "some_attribute", str, False, "Some attribute descriptions."
                    )
                ],
            ),
        )

        cls.list_of_messages = (
            DialogueMessage(
                OefSearchMessage.Performative.SEARCH_SERVICES, {"query": "some_query"}
            ),
        )

    def test_setup(self):
        """Test the setup method of the oef handler."""
        assert self.oef_search_handler.setup() is None
        self.assert_quantity_in_outbox(0)

    def test_handle_unidentified_dialogue(self):
        """Test the _handle_unidentified_dialogue method of the oef handler."""
        # setup
        incorrect_dialogue_reference = ("", "")
        incoming_message = self.build_incoming_message(
            message_type=OefSearchMessage,
            dialogue_reference=incorrect_dialogue_reference,
            performative=OefSearchMessage.Performative.SEARCH_RESULT,
            to=str(self.skill.skill_context.skill_id),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.oef_search_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.WARNING,
            f"received invalid oef_search message={incoming_message}, unidentified dialogue.",
        )

    def test_handle_error(self):
        """Test the _handle_error method of the oef handler."""
        # setup
        oef_dialogue = self.prepare_skill_dialogue(
            dialogues=self.oef_dialogues, messages=self.list_of_messages[:1],
        )
        incoming_message = cast(
            OefSearchMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=oef_dialogue,
                performative=OefSearchMessage.Performative.OEF_ERROR,
                to=str(self.skill.skill_context.skill_id),
                oef_error_operation=OefSearchMessage.OefErrorOperation.SEARCH_SERVICES,
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.oef_search_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.WARNING,
            f"received OEF Search error: dialogue_reference={oef_dialogue.dialogue_label.dialogue_reference}, oef_error_operation={incoming_message.oef_error_operation}",
        )

    def test_on_search_result_i(self):
        """Test the _on_search_result method of the oef handler."""
        # setup
        oef_dialogue = cast(
            OefSearchDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.oef_dialogues, messages=self.list_of_messages[:1],
            ),
        )
        oef_dialogue._is_seller_search = True
        search_for = "sellers"

        incoming_message = cast(
            OefSearchMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=oef_dialogue,
                performative=OefSearchMessage.Performative.SEARCH_RESULT,
                to=str(self.skill.skill_context.skill_id),
                agents=tuple(self.found_agents),
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            with patch.object(
                self.strategy, "get_own_services_query", return_value=self.cfp_query
            ) as mock_own:
                self.oef_search_handler.handle(incoming_message)

        # after
        self.assert_quantity_in_outbox(len(self.found_agents_less_self))

        # _handle_search
        mock_logger.assert_any_call(
            logging.INFO,
            f"found potential {search_for} agents={list(map(lambda x: x[-5:], self.found_agents_less_self))} on search_id={incoming_message.dialogue_reference[0]}.",
        )
        mock_own.assert_called_once()

        for agent in self.found_agents_less_self:
            mock_logger.assert_any_call(
                logging.INFO, f"sending CFP to agent={agent[-5:]}",
            )
            has_attributes, error_str = self.message_has_attributes(
                actual_message=self.get_message_from_outbox(),
                message_type=FipaMessage,
                performative=FipaMessage.Performative.CFP,
                to=agent,
                sender=self.self_address,
                query=self.cfp_query,
            )
            assert has_attributes, error_str

    def test_on_search_result_ii(self):
        """Test the _on_search_result method of the oef handler where number of agents found is 0."""
        # setup
        oef_dialogue = cast(
            OefSearchDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.oef_dialogues, messages=self.list_of_messages[:1],
            ),
        )
        oef_dialogue._is_seller_search = False
        search_for = "buyers"

        incoming_message = cast(
            OefSearchMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=oef_dialogue,
                performative=OefSearchMessage.Performative.SEARCH_RESULT,
                to=str(self.skill.skill_context.skill_id),
                agents=tuple(),
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.oef_search_handler.handle(incoming_message)

        # after
        self.assert_quantity_in_outbox(0)

        # _handle_search
        mock_logger.assert_any_call(
            logging.INFO,
            f"found no {search_for} agents on search_id={incoming_message.dialogue_reference[0]}, continue searching.",
        )

    def test_handle_invalid(self):
        """Test the _handle_invalid method of the oef handler."""
        # setup
        invalid_performative = OefSearchMessage.Performative.UNREGISTER_SERVICE
        incoming_message = self.build_incoming_message(
            message_type=OefSearchMessage,
            dialogue_reference=("1", ""),
            performative=invalid_performative,
            to=str(self.skill.skill_context.skill_id),
            service_description="some_service_description",
        )

        # operation
        with patch.object(self.oef_search_handler.context.logger, "log") as mock_logger:
            self.oef_search_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.WARNING,
            f"cannot handle oef_search message of performative={invalid_performative} in dialogue={self.oef_dialogues.get_dialogue(incoming_message)}.",
        )

    def test_teardown(self):
        """Test the teardown method of the oef_search handler."""
        assert self.oef_search_handler.teardown() is None
        self.assert_quantity_in_outbox(0)


class TestContractApiHandler(BaseSkillTestCase):
    """Test contract_api handler of tac negotiation."""

    path_to_skill = Path(ROOT_DIR, "packages", "fetchai", "skills", "tac_negotiation")
    is_agent_to_agent_messages = False

    @classmethod
    def setup(cls):
        """Setup the test class."""
        super().setup()
        cls.contract_api_handler = cast(
            ContractApiHandler, cls._skill.skill_context.handlers.contract_api
        )
        cls.logger = cls.contract_api_handler.context.logger

        cls.contract_api_dialogues = cast(
            ContractApiDialogues, cls._skill.skill_context.contract_api_dialogues
        )
        cls.signing_dialogues = cast(
            SigningDialogues, cls._skill.skill_context.signing_dialogues
        )
        cls.fipa_dialogues = cast(
            FipaDialogues, cls._skill.skill_context.fipa_dialogues
        )

        cls.ledger_id = "some_ledger_id"
        cls.contract_id = "some_contract_id"
        cls.contract_address = "some_contract_address"
        cls.callable = "some_callable"
        cls.kwargs = Kwargs({"some_key": "some_value"})
        cls.body = {"some_key": "some_value"}
        cls.body_bytes = b"some_body"
        cls.nonce = "some_nonce"
        cls.counterprty_address = COUNTERPARTY_AGENT_ADDRESS
        cls.amount_by_currency_id = {"1": 50}
        cls.quantities_by_good_id = {"2": -10}
        cls.terms = Terms(
            cls.ledger_id,
            cls._skill.skill_context.agent_address,
            cls.counterprty_address,
            cls.amount_by_currency_id,
            cls.quantities_by_good_id,
            cls.nonce,
        )

        cls.list_of_contract_api_messages_get_deploy_tx = (
            DialogueMessage(
                ContractApiMessage.Performative.GET_DEPLOY_TRANSACTION,
                {
                    "ledger_id": cls.ledger_id,
                    "contract_id": cls.contract_id,
                    "callable": cls.callable,
                    "kwargs": cls.kwargs,
                },
            ),
        )
        cls.list_of_contract_api_messages_raw_msg = (
            DialogueMessage(
                ContractApiMessage.Performative.GET_RAW_MESSAGE,
                {
                    "ledger_id": cls.ledger_id,
                    "contract_id": cls.contract_id,
                    "contract_address": cls.contract_address,
                    "callable": cls.callable,
                    "kwargs": cls.kwargs,
                },
            ),
        )

        cls.cfp_query = Query(
            [Constraint("some_attribute", ConstraintType("==", "some_service"))],
            DataModel(
                SUPPLY_DATAMODEL_NAME,
                [
                    Attribute(
                        "some_attribute", str, False, "Some attribute descriptions."
                    )
                ],
            ),
        )
        cls.list_of_fipa_messages = (
            DialogueMessage(FipaMessage.Performative.CFP, {"query": cls.cfp_query}),
        )

    def test_setup(self):
        """Test the setup method of the contract_api handler."""
        assert self.contract_api_handler.setup() is None
        self.assert_quantity_in_outbox(0)

    def test_handle_unidentified_dialogue(self):
        """Test the _handle_unidentified_dialogue method of the signing handler."""
        # setup
        incorrect_dialogue_reference = ("", "")
        incoming_message = self.build_incoming_message(
            message_type=ContractApiMessage,
            dialogue_reference=incorrect_dialogue_reference,
            performative=ContractApiMessage.Performative.STATE,
            state=State("some_ledger_id", {"some_key": "some_value"}),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.contract_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received invalid contract_api message={incoming_message}, unidentified dialogue.",
        )

    def test_handle_raw_message(self):
        """Test the _handle_raw_message method of the signing handler."""
        # setup
        fipa_dialogue = cast(
            FipaDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.fipa_dialogues, messages=self.list_of_fipa_messages[:1],
            ),
        )
        fipa_dialogue.terms = self.terms

        contract_api_dialogue = cast(
            ContractApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.contract_api_dialogues,
                messages=self.list_of_contract_api_messages_raw_msg[:1],
            ),
        )
        contract_api_dialogue.associated_fipa_dialogue = fipa_dialogue

        incoming_message = cast(
            ContractApiMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=contract_api_dialogue,
                performative=ContractApiMessage.Performative.RAW_MESSAGE,
                raw_message=ContractApiMessage.RawMessage(
                    self.ledger_id, self.body_bytes
                ),
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.contract_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO, f"received raw message={incoming_message}"
        )

        self.assert_quantity_in_decision_making_queue(1)
        message = self.get_message_from_decision_maker_inbox()
        has_attributes, error_str = self.message_has_attributes(
            actual_message=message,
            message_type=SigningMessage,
            performative=SigningMessage.Performative.SIGN_MESSAGE,
            to=self.skill.skill_context.decision_maker_address,
            sender=str(self.skill.skill_context.skill_id),
            terms=self.terms,
            raw_message=RawMessage(
                self.ledger_id, self.body_bytes, is_deprecated_mode=True,
            ),
        )
        assert has_attributes, error_str

        assert (
            cast(
                SigningDialogue, self.signing_dialogues.get_dialogue(message)
            ).associated_fipa_dialogue
            == fipa_dialogue
        )

        mock_logger.assert_any_call(
            logging.INFO,
            "proposing the message to the decision maker. Waiting for confirmation ...",
        )

    def test_handle_raw_transaction(self):
        """Test the _handle_signed_transaction method of the signing handler."""
        # setup
        fipa_dialogue = cast(
            FipaDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.fipa_dialogues, messages=self.list_of_fipa_messages[:1],
            ),
        )
        fipa_dialogue.terms = self.terms

        contract_api_dialogue = cast(
            ContractApiDialogue,
            self.prepare_skill_dialogue(
                dialogues=self.contract_api_dialogues,
                messages=self.list_of_contract_api_messages_get_deploy_tx[:1],
            ),
        )
        contract_api_dialogue.associated_fipa_dialogue = fipa_dialogue

        incoming_message = cast(
            ContractApiMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=contract_api_dialogue,
                performative=ContractApiMessage.Performative.RAW_TRANSACTION,
                raw_transaction=ContractApiMessage.RawTransaction(
                    self.ledger_id, self.body
                ),
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.contract_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO, f"received raw transaction={incoming_message}"
        )

        self.assert_quantity_in_decision_making_queue(1)
        message = self.get_message_from_decision_maker_inbox()
        has_attributes, error_str = self.message_has_attributes(
            actual_message=message,
            message_type=SigningMessage,
            performative=SigningMessage.Performative.SIGN_TRANSACTION,
            to=self.skill.skill_context.decision_maker_address,
            sender=str(self.skill.skill_context.skill_id),
            terms=self.terms,
            raw_transaction=incoming_message.raw_transaction,
        )
        assert has_attributes, error_str

        assert (
            cast(
                SigningDialogue, self.signing_dialogues.get_dialogue(message)
            ).associated_fipa_dialogue
            == fipa_dialogue
        )

        mock_logger.assert_any_call(
            logging.INFO,
            "proposing the transaction to the decision maker. Waiting for confirmation ...",
        )

    def test_handle_error(self):
        """Test the _handle_error method of the signing handler."""
        # setup
        contract_api_dialogue = self.prepare_skill_dialogue(
            dialogues=self.contract_api_dialogues,
            messages=self.list_of_contract_api_messages_get_deploy_tx[:1],
        )
        incoming_message = cast(
            ContractApiMessage,
            self.build_incoming_message_for_skill_dialogue(
                dialogue=contract_api_dialogue,
                performative=ContractApiMessage.Performative.ERROR,
                data=b"some_data",
            ),
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.contract_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.INFO,
            f"received contract_api error message={incoming_message} in dialogue={contract_api_dialogue}.",
        )

    def test_handle_invalid(self):
        """Test the _handle_invalid method of the signing handler."""
        # setup
        invalid_performative = ContractApiMessage.Performative.GET_DEPLOY_TRANSACTION
        incoming_message = self.build_incoming_message(
            message_type=ContractApiMessage,
            dialogue_reference=("1", ""),
            performative=invalid_performative,
            ledger_id=self.ledger_id,
            contract_id=self.contract_id,
            callable=self.callable,
            kwargs=self.kwargs,
        )

        # operation
        with patch.object(self.logger, "log") as mock_logger:
            self.contract_api_handler.handle(incoming_message)

        # after
        mock_logger.assert_any_call(
            logging.WARNING,
            f"cannot handle contract_api message of performative={invalid_performative} in dialogue={self.contract_api_dialogues.get_dialogue(incoming_message)}.",
        )

    def test_teardown(self):
        """Test the teardown method of the contract_api handler."""
        assert self.contract_api_handler.teardown() is None
        self.assert_quantity_in_outbox(0)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: acn_message.proto

import sys

_b = sys.version_info[0] < 3 and (lambda x: x) or (lambda x: x.encode("latin1"))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


DESCRIPTOR = _descriptor.FileDescriptor(
    name="acn_message.proto",
    package="dhtnode",
    syntax="proto3",
    serialized_pb=_b(
        '\n\x11\x61\x63n_message.proto\x12\x07\x64htnode"\xac\x01\n\x0b\x41gentRecord\x12\x12\n\nservice_id\x18\x01 \x01(\t\x12\x11\n\tledger_id\x18\x02 \x01(\t\x12\x0f\n\x07\x61\x64\x64ress\x18\x03 \x01(\t\x12\x12\n\npublic_key\x18\x04 \x01(\t\x12\x17\n\x0fpeer_public_key\x18\x05 \x01(\t\x12\x11\n\tsignature\x18\x06 \x01(\t\x12\x12\n\nnot_before\x18\x07 \x01(\t\x12\x11\n\tnot_after\x18\x08 \x01(\t"0\n\x08Register\x12$\n\x06record\x18\x01 \x01(\x0b\x32\x14.dhtnode.AgentRecord"&\n\rLookupRequest\x12\x15\n\ragent_address\x18\x01 \x01(\t"<\n\x0eLookupResponse\x12*\n\x0c\x61gent_record\x18\x01 \x01(\x0b\x32\x14.dhtnode.AgentRecord"B\n\x0b\x41\x65\x61\x45nvelope\x12\r\n\x05\x65nvel\x18\x01 \x01(\x0c\x12$\n\x06record\x18\x02 \x01(\x0b\x32\x14.dhtnode.AgentRecord"\xed\x02\n\x06Status\x12%\n\x04\x63ode\x18\x01 \x01(\x0e\x32\x17.dhtnode.Status.ErrCode\x12\x0c\n\x04msgs\x18\x02 \x03(\t"\xad\x02\n\x07\x45rrCode\x12\x0b\n\x07SUCCESS\x10\x00\x12\x1d\n\x19\x45RROR_UNSUPPORTED_VERSION\x10\x01\x12\x1c\n\x18\x45RROR_UNEXPECTED_PAYLOAD\x10\x02\x12\x11\n\rERROR_GENERIC\x10\x03\x12\x17\n\x13\x45RROR_SERIALIZATION\x10\x04\x12\x1d\n\x19\x45RROR_WRONG_AGENT_ADDRESS\x10\n\x12\x1a\n\x16\x45RROR_WRONG_PUBLIC_KEY\x10\x0b\x12\x17\n\x13\x45RROR_INVALID_PROOF\x10\x0c\x12\x1c\n\x18\x45RROR_UNSUPPORTED_LEGDER\x10\r\x12\x1f\n\x1b\x45RROR_UNKNOWN_AGENT_ADDRESS\x10\x14\x12\x19\n\x15\x45RROR_AGENT_NOT_READY\x10\x15"\x86\x02\n\nAcnMessage\x12\x0f\n\x07version\x18\x01 \x01(\t\x12!\n\x06status\x18\x02 \x01(\x0b\x32\x0f.dhtnode.StatusH\x00\x12%\n\x08register\x18\x03 \x01(\x0b\x32\x11.dhtnode.RegisterH\x00\x12\x30\n\x0elookup_request\x18\x04 \x01(\x0b\x32\x16.dhtnode.LookupRequestH\x00\x12\x32\n\x0flookup_response\x18\x05 \x01(\x0b\x32\x17.dhtnode.LookupResponseH\x00\x12,\n\x0c\x61\x65\x61_envelope\x18\x06 \x01(\x0b\x32\x14.dhtnode.AeaEnvelopeH\x00\x42\t\n\x07payloadb\x06proto3'
    ),
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)


_STATUS_ERRCODE = _descriptor.EnumDescriptor(
    name="ErrCode",
    full_name="dhtnode.Status.ErrCode",
    filename=None,
    file=DESCRIPTOR,
    values=[
        _descriptor.EnumValueDescriptor(
            name="SUCCESS", index=0, number=0, options=None, type=None
        ),
        _descriptor.EnumValueDescriptor(
            name="ERROR_UNSUPPORTED_VERSION", index=1, number=1, options=None, type=None
        ),
        _descriptor.EnumValueDescriptor(
            name="ERROR_UNEXPECTED_PAYLOAD", index=2, number=2, options=None, type=None
        ),
        _descriptor.EnumValueDescriptor(
            name="ERROR_GENERIC", index=3, number=3, options=None, type=None
        ),
        _descriptor.EnumValueDescriptor(
            name="ERROR_SERIALIZATION", index=4, number=4, options=None, type=None
        ),
        _descriptor.EnumValueDescriptor(
            name="ERROR_WRONG_AGENT_ADDRESS",
            index=5,
            number=10,
            options=None,
            type=None,
        ),
        _descriptor.EnumValueDescriptor(
            name="ERROR_WRONG_PUBLIC_KEY", index=6, number=11, options=None, type=None
        ),
        _descriptor.EnumValueDescriptor(
            name="ERROR_INVALID_PROOF", index=7, number=12, options=None, type=None
        ),
        _descriptor.EnumValueDescriptor(
            name="ERROR_UNSUPPORTED_LEGDER", index=8, number=13, options=None, type=None
        ),
        _descriptor.EnumValueDescriptor(
            name="ERROR_UNKNOWN_AGENT_ADDRESS",
            index=9,
            number=20,
            options=None,
            type=None,
        ),
        _descriptor.EnumValueDescriptor(
            name="ERROR_AGENT_NOT_READY", index=10, number=21, options=None, type=None
        ),
    ],
    containing_type=None,
    options=None,
    serialized_start=490,
    serialized_end=791,
)
_sym_db.RegisterEnumDescriptor(_STATUS_ERRCODE)


_AGENTRECORD = _descriptor.Descriptor(
    name="AgentRecord",
    full_name="dhtnode.AgentRecord",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="service_id",
            full_name="dhtnode.AgentRecord.service_id",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
        ),
        _descriptor.FieldDescriptor(
            name="ledger_id",
            full_name="dhtnode.AgentRecord.ledger_id",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
        ),
        _descriptor.FieldDescriptor(
            name="address",
            full_name="dhtnode.AgentRecord.address",
            index=2,
            number=3,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
        ),
        _descriptor.FieldDescriptor(
            name="public_key",
            full_name="dhtnode.AgentRecord.public_key",
            index=3,
            number=4,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
        ),
        _descriptor.FieldDescriptor(
            name="peer_public_key",
            full_name="dhtnode.AgentRecord.peer_public_key",
            index=4,
            number=5,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
        ),
        _descriptor.FieldDescriptor(
            name="signature",
            full_name="dhtnode.AgentRecord.signature",
            index=5,
            number=6,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
        ),
        _descriptor.FieldDescriptor(
            name="not_before",
            full_name="dhtnode.AgentRecord.not_before",
            index=6,
            number=7,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
        ),
        _descriptor.FieldDescriptor(
            name="not_after",
            full_name="dhtnode.AgentRecord.not_after",
            index=7,
            number=8,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=31,
    serialized_end=203,
)


_REGISTER = _descriptor.Descriptor(
    name="Register",
    full_name="dhtnode.Register",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="record",
            full_name="dhtnode.Register.record",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=205,
    serialized_end=253,
)


_LOOKUPREQUEST = _descriptor.Descriptor(
    name="LookupRequest",
    full_name="dhtnode.LookupRequest",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="agent_address",
            full_name="dhtnode.LookupRequest.agent_address",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=255,
    serialized_end=293,
)


_LOOKUPRESPONSE = _descriptor.Descriptor(
    name="LookupResponse",
    full_name="dhtnode.LookupResponse",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="agent_record",
            full_name="dhtnode.LookupResponse.agent_record",
            index=0,
            number=1,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=295,
    serialized_end=355,
)


_AEAENVELOPE = _descriptor.Descriptor(
    name="AeaEnvelope",
    full_name="dhtnode.AeaEnvelope",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="envel",
            full_name="dhtnode.AeaEnvelope.envel",
            index=0,
            number=1,
            type=12,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b(""),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
        ),
        _descriptor.FieldDescriptor(
            name="record",
            full_name="dhtnode.AeaEnvelope.record",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=357,
    serialized_end=423,
)


_STATUS = _descriptor.Descriptor(
    name="Status",
    full_name="dhtnode.Status",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="code",
            full_name="dhtnode.Status.code",
            index=0,
            number=1,
            type=14,
            cpp_type=8,
            label=1,
            has_default_value=False,
            default_value=0,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
        ),
        _descriptor.FieldDescriptor(
            name="msgs",
            full_name="dhtnode.Status.msgs",
            index=1,
            number=2,
            type=9,
            cpp_type=9,
            label=3,
            has_default_value=False,
            default_value=[],
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[_STATUS_ERRCODE,],
    options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[],
    serialized_start=426,
    serialized_end=791,
)


_ACNMESSAGE = _descriptor.Descriptor(
    name="AcnMessage",
    full_name="dhtnode.AcnMessage",
    filename=None,
    file=DESCRIPTOR,
    containing_type=None,
    fields=[
        _descriptor.FieldDescriptor(
            name="version",
            full_name="dhtnode.AcnMessage.version",
            index=0,
            number=1,
            type=9,
            cpp_type=9,
            label=1,
            has_default_value=False,
            default_value=_b("").decode("utf-8"),
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
        ),
        _descriptor.FieldDescriptor(
            name="status",
            full_name="dhtnode.AcnMessage.status",
            index=1,
            number=2,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
        ),
        _descriptor.FieldDescriptor(
            name="register",
            full_name="dhtnode.AcnMessage.register",
            index=2,
            number=3,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
        ),
        _descriptor.FieldDescriptor(
            name="lookup_request",
            full_name="dhtnode.AcnMessage.lookup_request",
            index=3,
            number=4,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
        ),
        _descriptor.FieldDescriptor(
            name="lookup_response",
            full_name="dhtnode.AcnMessage.lookup_response",
            index=4,
            number=5,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
        ),
        _descriptor.FieldDescriptor(
            name="aea_envelope",
            full_name="dhtnode.AcnMessage.aea_envelope",
            index=5,
            number=6,
            type=11,
            cpp_type=10,
            label=1,
            has_default_value=False,
            default_value=None,
            message_type=None,
            enum_type=None,
            containing_type=None,
            is_extension=False,
            extension_scope=None,
            options=None,
        ),
    ],
    extensions=[],
    nested_types=[],
    enum_types=[],
    options=None,
    is_extendable=False,
    syntax="proto3",
    extension_ranges=[],
    oneofs=[
        _descriptor.OneofDescriptor(
            name="payload",
            full_name="dhtnode.AcnMessage.payload",
            index=0,
            containing_type=None,
            fields=[],
        ),
    ],
    serialized_start=794,
    serialized_end=1056,
)

_REGISTER.fields_by_name["record"].message_type = _AGENTRECORD
_LOOKUPRESPONSE.fields_by_name["agent_record"].message_type = _AGENTRECORD
_AEAENVELOPE.fields_by_name["record"].message_type = _AGENTRECORD
_STATUS.fields_by_name["code"].enum_type = _STATUS_ERRCODE
_STATUS_ERRCODE.containing_type = _STATUS
_ACNMESSAGE.fields_by_name["status"].message_type = _STATUS
_ACNMESSAGE.fields_by_name["register"].message_type = _REGISTER
_ACNMESSAGE.fields_by_name["lookup_request"].message_type = _LOOKUPREQUEST
_ACNMESSAGE.fields_by_name["lookup_response"].message_type = _LOOKUPRESPONSE
_ACNMESSAGE.fields_by_name["aea_envelope"].message_type = _AEAENVELOPE
_ACNMESSAGE.oneofs_by_name["payload"].fields.append(
    _ACNMESSAGE.fields_by_name["status"]
)
_ACNMESSAGE.fields_by_name["status"].containing_oneof = _ACNMESSAGE.oneofs_by_name[
    "payload"
]
_ACNMESSAGE.oneofs_by_name["payload"].fields.append(
    _ACNMESSAGE.fields_by_name["register"]
)
_ACNMESSAGE.fields_by_name["register"].containing_oneof = _ACNMESSAGE.oneofs_by_name[
    "payload"
]
_ACNMESSAGE.oneofs_by_name["payload"].fields.append(
    _ACNMESSAGE.fields_by_name["lookup_request"]
)
_ACNMESSAGE.fields_by_name[
    "lookup_request"
].containing_oneof = _ACNMESSAGE.oneofs_by_name["payload"]
_ACNMESSAGE.oneofs_by_name["payload"].fields.append(
    _ACNMESSAGE.fields_by_name["lookup_response"]
)
_ACNMESSAGE.fields_by_name[
    "lookup_response"
].containing_oneof = _ACNMESSAGE.oneofs_by_name["payload"]
_ACNMESSAGE.oneofs_by_name["payload"].fields.append(
    _ACNMESSAGE.fields_by_name["aea_envelope"]
)
_ACNMESSAGE.fields_by_name[
    "aea_envelope"
].containing_oneof = _ACNMESSAGE.oneofs_by_name["payload"]
DESCRIPTOR.message_types_by_name["AgentRecord"] = _AGENTRECORD
DESCRIPTOR.message_types_by_name["Register"] = _REGISTER
DESCRIPTOR.message_types_by_name["LookupRequest"] = _LOOKUPREQUEST
DESCRIPTOR.message_types_by_name["LookupResponse"] = _LOOKUPRESPONSE
DESCRIPTOR.message_types_by_name["AeaEnvelope"] = _AEAENVELOPE
DESCRIPTOR.message_types_by_name["Status"] = _STATUS
DESCRIPTOR.message_types_by_name["AcnMessage"] = _ACNMESSAGE

AgentRecord = _reflection.GeneratedProtocolMessageType(
    "AgentRecord",
    (_message.Message,),
    dict(
        DESCRIPTOR=_AGENTRECORD,
        __module__="acn_message_pb2"
        # @@protoc_insertion_point(class_scope:dhtnode.AgentRecord)
    ),
)
_sym_db.RegisterMessage(AgentRecord)

Register = _reflection.GeneratedProtocolMessageType(
    "Register",
    (_message.Message,),
    dict(
        DESCRIPTOR=_REGISTER,
        __module__="acn_message_pb2"
        # @@protoc_insertion_point(class_scope:dhtnode.Register)
    ),
)
_sym_db.RegisterMessage(Register)

LookupRequest = _reflection.GeneratedProtocolMessageType(
    "LookupRequest",
    (_message.Message,),
    dict(
        DESCRIPTOR=_LOOKUPREQUEST,
        __module__="acn_message_pb2"
        # @@protoc_insertion_point(class_scope:dhtnode.LookupRequest)
    ),
)
_sym_db.RegisterMessage(LookupRequest)

LookupResponse = _reflection.GeneratedProtocolMessageType(
    "LookupResponse",
    (_message.Message,),
    dict(
        DESCRIPTOR=_LOOKUPRESPONSE,
        __module__="acn_message_pb2"
        # @@protoc_insertion_point(class_scope:dhtnode.LookupResponse)
    ),
)
_sym_db.RegisterMessage(LookupResponse)

AeaEnvelope = _reflection.GeneratedProtocolMessageType(
    "AeaEnvelope",
    (_message.Message,),
    dict(
        DESCRIPTOR=_AEAENVELOPE,
        __module__="acn_message_pb2"
        # @@protoc_insertion_point(class_scope:dhtnode.AeaEnvelope)
    ),
)
_sym_db.RegisterMessage(AeaEnvelope)

Status = _reflection.GeneratedProtocolMessageType(
    "Status",
    (_message.Message,),
    dict(
        DESCRIPTOR=_STATUS,
        __module__="acn_message_pb2"
        # @@protoc_insertion_point(class_scope:dhtnode.Status)
    ),
)
_sym_db.RegisterMessage(Status)

AcnMessage = _reflection.GeneratedProtocolMessageType(
    "AcnMessage",
    (_message.Message,),
    dict(
        DESCRIPTOR=_ACNMESSAGE,
        __module__="acn_message_pb2"
        # @@protoc_insertion_point(class_scope:dhtnode.AcnMessage)
    ),
)
_sym_db.RegisterMessage(AcnMessage)


# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2018-2020 Fetch.AI Limited
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""This test module contains the tests for commands in aea.cli.transfer module."""
import random
import string
from pathlib import Path
from unittest.mock import patch

import pytest
from aea_ledger_cosmos import CosmosCrypto
from aea_ledger_fetchai import FetchAICrypto
from click.exceptions import ClickException

from aea.cli.transfer import wait_tx_settled
from aea.cli.utils.package_utils import try_get_balance
from aea.configurations.manager import AgentConfigManager
from aea.crypto.helpers import (
    get_wallet_from_agent_config,
    private_key_verify_or_create,
)
from aea.helpers.base import cd
from aea.test_tools.test_cases import AEATestCaseEmpty

from tests.common.utils import wait_for_condition
from tests.conftest import MAX_FLAKY_RERUNS


class TestCliTransferFetchAINetwork(AEATestCaseEmpty):
    """Test cli transfer command."""

    LEDGER_ID = FetchAICrypto.identifier
    ANOTHER_LEDGER_ID = CosmosCrypto.identifier

    @classmethod
    def setup_class(cls):
        """Set up the test class."""
        super(TestCliTransferFetchAINetwork, cls).setup_class()
        cls.agent_name2 = "agent-" + "".join(
            random.choices(string.ascii_lowercase, k=5)
        )
        cls.create_agents(cls.agent_name2)

        cls.gen_key(cls.agent_name)
        cls.gen_key(cls.agent_name2)

    @classmethod
    def gen_key(cls, agent_name: str) -> None:
        """Generate crypto key."""
        cls.set_agent_context(agent_name)
        key_file = f"{cls.LEDGER_ID}.key"
        assert cls.run_cli_command(
            "generate-key", cls.LEDGER_ID, key_file, cwd=cls._get_cwd()
        )
        assert cls.run_cli_command(
            "add-key", cls.LEDGER_ID, key_file, cwd=cls._get_cwd()
        )

    def get_address(self) -> str:
        """Get current agent address."""
        result = self.invoke("get-address", self.LEDGER_ID)
        return result.stdout_bytes.decode("utf-8").strip()

    def get_balance(self) -> int:
        """Get balance for current agent."""
        with cd(self._get_cwd()):
            agent_config = AgentConfigManager.verify_or_create_private_keys(
                Path("."),
                substitude_env_vars=False,
                private_key_helper=private_key_verify_or_create,
            ).agent_config
            wallet = get_wallet_from_agent_config(agent_config)
            return int(try_get_balance(agent_config, wallet, self.LEDGER_ID))

    @pytest.mark.flaky(reruns=MAX_FLAKY_RERUNS)
    def test_integration(self):
        """Perform integration tests of cli transfer command with real transfer."""
        self.set_agent_context(self.agent_name2)
        agent2_balance = self.get_balance()
        agent2_address = self.get_address()
        assert agent2_balance == 0

        self.set_agent_context(self.agent_name)
        self.generate_wealth()

        wait_for_condition(lambda: self.get_balance() > 0, timeout=15, period=1)

        agent1_balance = self.get_balance()
        assert agent1_balance > 0

        amount = round(agent1_balance / 10)
        fee = round(agent1_balance / 20)

        self.invoke(
            "transfer", self.LEDGER_ID, agent2_address, str(amount), str(fee), "-y"
        )

        wait_for_condition(
            lambda: self.get_balance() == (agent1_balance - amount - fee),
            timeout=15,
            period=1,
        )

        self.set_agent_context(self.agent_name2)
        wait_for_condition(
            lambda: self.get_balance() == (agent2_balance + amount),
            timeout=15,
            period=1,
        )

    @patch("aea.cli.transfer.do_transfer", return_value="some_digest")
    @patch("aea.cli.transfer.click.confirm", return_value=None)
    @patch("aea.cli.transfer.wait_tx_settled", return_value=None)
    def test_yes_option_enabled(
        self, wait_tx_settled_mock, confirm_mock, do_transfer_mock
    ):
        """Test yes option is enabled."""
        self.invoke(
            "transfer", self.LEDGER_ID, self.get_address(), "100000", "100", "-y"
        )
        confirm_mock.assert_not_called()

    @patch("aea.cli.transfer.do_transfer", return_value="some_digest")
    @patch("aea.cli.transfer.click.confirm", return_value=None)
    @patch("aea.cli.transfer.wait_tx_settled", return_value=None)
    def test_yes_option_disabled(
        self, wait_tx_settled_mock, confirm_mock, do_transfer_mock
    ):
        """Test yes option is disabled."""
        self.invoke("transfer", self.LEDGER_ID, self.get_address(), "100000", "100")
        confirm_mock.assert_called_once()

    @patch("aea.cli.transfer.do_transfer", return_value="some_digest")
    @patch("aea.cli.transfer.click.confirm", return_value=None)
    @patch("aea.cli.transfer.wait_tx_settled", return_value=None)
    def test_sync_option_enabled(
        self, wait_tx_settled_mock, confirm_mock, do_transfer_mock
    ):
        """Test sync option is enabled."""
        self.invoke(
            "transfer", self.LEDGER_ID, self.get_address(), "100000", "100", "-y"
        )
        wait_tx_settled_mock.assert_not_called()

    @patch("aea.cli.transfer.do_transfer", return_value="some_digest")
    @patch("aea.cli.transfer.click.confirm", return_value=None)
    @patch("aea.cli.transfer.wait_tx_settled", return_value=None)
    def test_sync_option_disabled(
        self, wait_tx_settled_mock, confirm_mock, do_transfer_mock
    ):
        """Test sync option is disabled."""
        self.invoke(
            "transfer",
            self.LEDGER_ID,
            self.get_address(),
            "100000",
            "100",
            "-y",
            "--sync",
        )
        wait_tx_settled_mock.assert_called_once()

    @patch("aea.cli.transfer.do_transfer", return_value=None)
    @patch("aea.cli.transfer.click.confirm", return_value=None)
    @patch("aea.cli.transfer.wait_tx_settled", return_value=None)
    def test_failed_on_send(self, wait_tx_settled_mock, confirm_mock, do_transfer_mock):
        """Test fail to send a transaction."""
        with pytest.raises(ClickException, match=r"Failed to send a transaction!"):
            self.invoke("transfer", self.LEDGER_ID, self.get_address(), "100000", "100")

    @patch("aea.cli.transfer.click.confirm", return_value=None)
    @patch("aea.cli.transfer.wait_tx_settled", return_value=None)
    def test_no_wallet_registered(self, wait_tx_settled_mock, confirm_mock):
        """Test no wallet for crypto id registered."""
        with pytest.raises(
            ClickException, match=r"No private key registered for `.*` in wallet!"
        ):
            self.invoke(
                "transfer", self.ANOTHER_LEDGER_ID, self.get_address(), "100000", "100"
            )

    @patch("aea.cli.transfer.try_get_balance", return_value=10)
    @patch("aea.cli.transfer.click.confirm", return_value=None)
    @patch("aea.cli.transfer.wait_tx_settled", return_value=None)
    def test_balance_too_low(
        self, wait_tx_settled_mock, confirm_mock, do_transfer_mock
    ):
        """Test balance too low exception."""
        with pytest.raises(
            ClickException,
            match=r"Balance is not enough! Available=[0-9]+, required=[0-9]+!",
        ):
            self.invoke("transfer", self.LEDGER_ID, self.get_address(), "100000", "100")

    @patch(
        "aea.cli.transfer.LedgerApis.is_transaction_settled", side_effects=[False, True]
    )
    def test_wait_tx_settled_ok(self, is_transaction_settled_mock):
        """Test wait tx settle is ok."""
        wait_tx_settled("some", "some", timeout=4)

    @patch("aea.cli.transfer.LedgerApis.is_transaction_settled", return_value=False)
    def test_wait_tx_settled_timeout(self, is_transaction_settled_mock):
        """Test wait tx settle fails with timeout error."""
        with pytest.raises(TimeoutError):
            wait_tx_settled("some", "some", timeout=0.5)
